import pathlib
import shutil
import tempfile
import unittest
from io import StringIO

from fontTools.voltLib.voltToFea import VoltToFea

DATADIR = pathlib.Path(__file__).parent / "data"


class ToFeaTest(unittest.TestCase):
    @classmethod
    def setup_class(cls):
        cls.tempdir = None
        cls.num_tempfiles = 0

    @classmethod
    def teardown_class(cls):
        if cls.tempdir:
            shutil.rmtree(cls.tempdir, ignore_errors=True)

    @classmethod
    def temp_path(cls):
        if not cls.tempdir:
            cls.tempdir = pathlib.Path(tempfile.mkdtemp())
        cls.num_tempfiles += 1
        return cls.tempdir / f"tmp{cls.num_tempfiles}"

    def test_def_glyph_base(self):
        fea = self.parse('DEF_GLYPH ".notdef" ID 0 TYPE BASE END_GLYPH')
        self.assertEqual(
            fea,
            "@GDEF_base = [.notdef];\n"
            "table GDEF {\n"
            "    GlyphClassDef @GDEF_base, , , ;\n"
            "} GDEF;\n",
        )

    def test_def_glyph_base_2_components(self):
        fea = self.parse(
            'DEF_GLYPH "glyphBase" ID 320 TYPE BASE COMPONENTS 2 END_GLYPH'
        )
        self.assertEqual(
            fea,
            "@GDEF_base = [glyphBase];\n"
            "table GDEF {\n"
            "    GlyphClassDef @GDEF_base, , , ;\n"
            "} GDEF;\n",
        )

    def test_def_glyph_ligature_2_components(self):
        fea = self.parse('DEF_GLYPH "f_f" ID 320 TYPE LIGATURE COMPONENTS 2 END_GLYPH')
        self.assertEqual(
            fea,
            "@GDEF_ligature = [f_f];\n"
            "table GDEF {\n"
            "    GlyphClassDef , @GDEF_ligature, , ;\n"
            "} GDEF;\n",
        )

    def test_def_glyph_mark(self):
        fea = self.parse('DEF_GLYPH "brevecomb" ID 320 TYPE MARK END_GLYPH')
        self.assertEqual(
            fea,
            "@GDEF_mark = [brevecomb];\n"
            "table GDEF {\n"
            "    GlyphClassDef , , @GDEF_mark, ;\n"
            "} GDEF;\n",
        )

    def test_def_glyph_component(self):
        fea = self.parse('DEF_GLYPH "f.f_f" ID 320 TYPE COMPONENT END_GLYPH')
        self.assertEqual(
            fea,
            "@GDEF_component = [f.f_f];\n"
            "table GDEF {\n"
            "    GlyphClassDef , , , @GDEF_component;\n"
            "} GDEF;\n",
        )

    def test_def_glyph_no_type(self):
        fea = self.parse('DEF_GLYPH "glyph20" ID 20 END_GLYPH')
        self.assertEqual(fea, "")

    def test_def_glyph_case_sensitive(self):
        fea = self.parse(
            'DEF_GLYPH "A" ID 3 UNICODE 65 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "a" ID 4 UNICODE 97 TYPE BASE END_GLYPH\n'
        )
        self.assertEqual(
            fea,
            "@GDEF_base = [A a];\n"
            "table GDEF {\n"
            "    GlyphClassDef @GDEF_base, , , ;\n"
            "} GDEF;\n",
        )

    def test_def_group_glyphs(self):
        fea = self.parse(
            'DEF_GROUP "aaccented"\n'
            'ENUM GLYPH "aacute" GLYPH "abreve" GLYPH "acircumflex" '
            'GLYPH "adieresis" GLYPH "ae" GLYPH "agrave" GLYPH "amacron" '
            'GLYPH "aogonek" GLYPH "aring" GLYPH "atilde" END_ENUM\n'
            "END_GROUP\n"
        )
        self.assertEqual(
            fea,
            "# Glyph classes\n"
            "@aaccented = [aacute abreve acircumflex adieresis ae"
            " agrave amacron aogonek aring atilde];",
        )

    def test_def_group_groups(self):
        fea = self.parse(
            'DEF_GROUP "Group1"\n'
            'ENUM GLYPH "a" GLYPH "b" GLYPH "c" GLYPH "d" END_ENUM\n'
            "END_GROUP\n"
            'DEF_GROUP "Group2"\n'
            'ENUM GLYPH "e" GLYPH "f" GLYPH "g" GLYPH "h" END_ENUM\n'
            "END_GROUP\n"
            'DEF_GROUP "TestGroup"\n'
            'ENUM GROUP "Group1" GROUP "Group2" END_ENUM\n'
            "END_GROUP\n"
        )
        self.assertEqual(
            fea,
            "# Glyph classes\n"
            "@Group1 = [a b c d];\n"
            "@Group2 = [e f g h];\n"
            "@TestGroup = [@Group1 @Group2];",
        )

    def test_def_group_groups_not_yet_defined(self):
        fea = self.parse(
            'DEF_GROUP "Group1"\n'
            'ENUM GLYPH "a" GLYPH "b" GLYPH "c" GLYPH "d" END_ENUM\n'
            "END_GROUP\n"
            'DEF_GROUP "TestGroup1"\n'
            'ENUM GROUP "Group1" GROUP "Group2" END_ENUM\n'
            "END_GROUP\n"
            'DEF_GROUP "TestGroup2"\n'
            'ENUM GROUP "Group2" END_ENUM\n'
            "END_GROUP\n"
            'DEF_GROUP "TestGroup3"\n'
            'ENUM GROUP "Group2" GROUP "Group1" END_ENUM\n'
            "END_GROUP\n"
            'DEF_GROUP "Group2"\n'
            'ENUM GLYPH "e" GLYPH "f" GLYPH "g" GLYPH "h" END_ENUM\n'
            "END_GROUP\n"
        )
        self.assertEqual(
            fea,
            "# Glyph classes\n"
            "@Group1 = [a b c d];\n"
            "@Group2 = [e f g h];\n"
            "@TestGroup1 = [@Group1 @Group2];\n"
            "@TestGroup2 = [@Group2];\n"
            "@TestGroup3 = [@Group2 @Group1];",
        )

    def test_def_group_glyphs_and_group(self):
        fea = self.parse(
            'DEF_GROUP "aaccented"\n'
            'ENUM GLYPH "aacute" GLYPH "abreve" GLYPH "acircumflex" '
            'GLYPH "adieresis" GLYPH "ae" GLYPH "agrave" GLYPH "amacron" '
            'GLYPH "aogonek" GLYPH "aring" GLYPH "atilde" END_ENUM\n'
            "END_GROUP\n"
            'DEF_GROUP "KERN_lc_a_2ND"\n'
            'ENUM GLYPH "a" GROUP "aaccented" END_ENUM\n'
            "END_GROUP"
        )
        self.assertEqual(
            fea,
            "# Glyph classes\n"
            "@aaccented = [aacute abreve acircumflex adieresis ae"
            " agrave amacron aogonek aring atilde];\n"
            "@KERN_lc_a_2ND = [a @aaccented];",
        )

    def test_def_group_range(self):
        fea = self.parse(
            'DEF_GLYPH "a" ID 163 UNICODE 97 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "agrave" ID 194 UNICODE 224 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "aacute" ID 195 UNICODE 225 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "acircumflex" ID 196 UNICODE 226 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "atilde" ID 197 UNICODE 227 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "c" ID 165 UNICODE 99 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "ccaron" ID 209 UNICODE 269 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "ccedilla" ID 210 UNICODE 231 TYPE BASE END_GLYPH\n'
            'DEF_GLYPH "cdotaccent" ID 210 UNICODE 267 TYPE BASE END_GLYPH\n'
            'DEF_GROUP "KERN_lc_a_2ND"\n'
            'ENUM RANGE "a" TO "atilde" GLYPH "b" RANGE "c" TO "cdotaccent" '
            "END_ENUM\n"
            "END_GROUP"
        )
        self.assertEqual(
            fea,
            "# Glyph classes\n"
            "@KERN_lc_a_2ND = [a - atilde b c - cdotaccent];\n"
            "@GDEF_base = [a agrave aacute acircumflex atilde c"
            " ccaron ccedilla cdotaccent];\n"
            "table GDEF {\n"
            "    GlyphClassDef @GDEF_base, , , ;\n"
            "} GDEF;\n",
        )

    def test_script_without_langsys(self):
        fea = self.parse('DEF_SCRIPT NAME "Latin" TAG "latn"\n' "END_SCRIPT")
        self.assertEqual(fea, "")

    def test_langsys_normal(self):
        fea = self.parse(
            'DEF_SCRIPT NAME "Latin" TAG "latn"\n'
            'DEF_LANGSYS NAME "Romanian" TAG "ROM "\n'
            "END_LANGSYS\n"
            'DEF_LANGSYS NAME "Moldavian" TAG "MOL "\n'
            "END_LANGSYS\n"
            "END_SCRIPT"
        )
        self.assertEqual(fea, "")

    def test_langsys_no_script_name(self):
        fea = self.parse(
            'DEF_SCRIPT TAG "latn"\n'
            'DEF_LANGSYS NAME "Default" TAG "dflt"\n'
            "END_LANGSYS\n"
            "END_SCRIPT"
        )
        self.assertEqual(fea, "")

    def test_langsys_lang_in_separate_scripts(self):
        fea = self.parse(
            'DEF_SCRIPT NAME "Default" TAG "DFLT"\n'
            'DEF_LANGSYS NAME "Default" TAG "dflt"\n'
            "END_LANGSYS\n"
            'DEF_LANGSYS NAME "Default" TAG "ROM "\n'
            "END_LANGSYS\n"
            "END_SCRIPT\n"
            'DEF_SCRIPT NAME "Latin" TAG "latn"\n'
            'DEF_LANGSYS NAME "Default" TAG "dflt"\n'
            "END_LANGSYS\n"
            'DEF_LANGSYS NAME "Default" TAG "ROM "\n'
            "END_LANGSYS\n"
            "END_SCRIPT"
        )
        self.assertEqual(fea, "")

    def test_langsys_no_lang_name(self):
        fea = self.parse(
            'DEF_SCRIPT NAME "Latin" TAG "latn"\n'
            'DEF_LANGSYS TAG "dflt"\n'
            "END_LANGSYS\n"
            "END_SCRIPT"
        )
        self.assertEqual(fea, "")

    def test_feature(self):
        fea = self.parse(
            'DEF_SCRIPT NAME "Latin" TAG "latn"\n'
            'DEF_LANGSYS NAME "Romanian" TAG "ROM "\n'
            'DEF_FEATURE NAME "Fractions" TAG "frac"\n'
            'LOOKUP "fraclookup"\n'
            "END_FEATURE\n"
            "END_LANGSYS\n"
            "END_SCRIPT\n"
            'DEF_LOOKUP "fraclookup" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "one" GLYPH "slash" GLYPH "two"\n'
            'WITH GLYPH "one_slash_two.frac"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        )
        self.assertEqual(
            fea,
            "\n# Lookups\n"
            "lookup fraclookup {\n"
            "    sub one slash two by one_slash_two.frac;\n"
            "} fraclookup;\n"
            "\n"
            "# Features\n"
            "feature frac {\n"
            "    script latn;\n"
            "    language ROM exclude_dflt;\n"
            "    lookup fraclookup;\n"
            "} frac;\n",
        )

    def test_feature_sub_lookups(self):
        fea = self.parse(
            'DEF_SCRIPT NAME "Latin" TAG "latn"\n'
            'DEF_LANGSYS NAME "Romanian" TAG "ROM "\n'
            'DEF_FEATURE NAME "Fractions" TAG "frac"\n'
            'LOOKUP "fraclookup\\1"\n'
            'LOOKUP "fraclookup\\1"\n'
            "END_FEATURE\n"
            "END_LANGSYS\n"
            "END_SCRIPT\n"
            'DEF_LOOKUP "fraclookup\\1" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION RTL\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "one" GLYPH "slash" GLYPH "two"\n'
            'WITH GLYPH "one_slash_two.frac"\n'
            "END_SUB\n"
            "END_SUBSTITUTION\n"
            'DEF_LOOKUP "fraclookup\\2" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION RTL\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "one" GLYPH "slash" GLYPH "three"\n'
            'WITH GLYPH "one_slash_three.frac"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        )
        self.assertEqual(
            fea,
            "\n# Lookups\n"
            "lookup fraclookup {\n"
            "    lookupflag RightToLeft;\n"
            "    # fraclookup\\1\n"
            "    sub one slash two by one_slash_two.frac;\n"
            "    subtable;\n"
            "    # fraclookup\\2\n"
            "    sub one slash three by one_slash_three.frac;\n"
            "} fraclookup;\n"
            "\n"
            "# Features\n"
            "feature frac {\n"
            "    script latn;\n"
            "    language ROM exclude_dflt;\n"
            "    lookup fraclookup;\n"
            "} frac;\n",
        )

    def test_lookup_comment(self):
        fea = self.parse(
            'DEF_LOOKUP "smcp" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            'COMMENTS "Smallcaps lookup for testing"\n'
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "a"\n'
            'WITH GLYPH "a.sc"\n'
            "END_SUB\n"
            'SUB GLYPH "b"\n'
            'WITH GLYPH "b.sc"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        )
        self.assertEqual(
            fea,
            "\n# Lookups\n"
            "lookup smcp {\n"
            "    # Smallcaps lookup for testing\n"
            "    sub a by a.sc;\n"
            "    sub b by b.sc;\n"
            "} smcp;\n",
        )

    def test_substitution_single(self):
        fea = self.parse(
            'DEF_LOOKUP "smcp" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "a"\n'
            'WITH GLYPH "a.sc"\n'
            "END_SUB\n"
            'SUB GLYPH "b"\n'
            'WITH GLYPH "b.sc"\n'
            "END_SUB\n"
            "SUB WITH\n"  # Empty substitution, will be ignored
            "END_SUB\n"
            "END_SUBSTITUTION"
        )
        self.assertEqual(
            fea,
            "\n# Lookups\n"
            "lookup smcp {\n"
            "    sub a by a.sc;\n"
            "    sub b by b.sc;\n"
            "} smcp;\n",
        )

    def test_substitution_single_in_context(self):
        fea = self.parse(
            'DEF_GROUP "Denominators" ENUM GLYPH "one.dnom" GLYPH "two.dnom" '
            "END_ENUM END_GROUP\n"
            'DEF_LOOKUP "fracdnom" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            'IN_CONTEXT LEFT ENUM GROUP "Denominators" GLYPH "fraction" '
            "END_ENUM\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "one"\n'
            'WITH GLYPH "one.dnom"\n'
            "END_SUB\n"
            'SUB GLYPH "two"\n'
            'WITH GLYPH "two.dnom"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        )
        self.assertEqual(
            fea,
            "# Glyph classes\n"
            "@Denominators = [one.dnom two.dnom];\n"
            "\n"
            "# Lookups\n"
            "lookup fracdnom {\n"
            "    sub [@Denominators fraction] one' by one.dnom;\n"
            "    sub [@Denominators fraction] two' by two.dnom;\n"
            "} fracdnom;\n",
        )

    def test_substitution_single_in_contexts(self):
        fea = self.parse(
            'DEF_GROUP "Hebrew" ENUM GLYPH "uni05D0" GLYPH "uni05D1" '
            "END_ENUM END_GROUP\n"
            'DEF_LOOKUP "HebrewCurrency" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            'RIGHT GROUP "Hebrew"\n'
            'RIGHT GLYPH "one.Hebr"\n'
            "END_CONTEXT\n"
            "IN_CONTEXT\n"
            'LEFT GROUP "Hebrew"\n'
            'LEFT GLYPH "one.Hebr"\n'
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "dollar"\n'
            'WITH GLYPH "dollar.Hebr"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        )
        self.assertEqual(
            fea,
            "# Glyph classes\n"
            "@Hebrew = [uni05D0 uni05D1];\n"
            "\n"
            "# Lookups\n"
            "lookup HebrewCurrency {\n"
            "    sub dollar' @Hebrew one.Hebr by dollar.Hebr;\n"
            "    sub @Hebrew one.Hebr dollar' by dollar.Hebr;\n"
            "} HebrewCurrency;\n",
        )

    def test_substitution_single_except_context(self):
        fea = self.parse(
            'DEF_GROUP "Hebrew" ENUM GLYPH "uni05D0" GLYPH "uni05D1" '
            "END_ENUM END_GROUP\n"
            'DEF_LOOKUP "HebrewCurrency" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "EXCEPT_CONTEXT\n"
            'RIGHT GROUP "Hebrew"\n'
            'RIGHT GLYPH "one.Hebr"\n'
            "END_CONTEXT\n"
            "IN_CONTEXT\n"
            'LEFT GROUP "Hebrew"\n'
            'LEFT GLYPH "one.Hebr"\n'
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "dollar"\n'
            'WITH GLYPH "dollar.Hebr"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        )
        self.assertEqual(
            fea,
            "# Glyph classes\n"
            "@Hebrew = [uni05D0 uni05D1];\n"
            "\n"
            "# Lookups\n"
            "lookup HebrewCurrency {\n"
            "    ignore sub dollar' @Hebrew one.Hebr;\n"
            "    sub @Hebrew one.Hebr dollar' by dollar.Hebr;\n"
            "} HebrewCurrency;\n",
        )

    def test_substitution_skip_base(self):
        fea = self.parse(
            'DEF_GROUP "SomeMarks" ENUM GLYPH "marka" GLYPH "markb" '
            "END_ENUM END_GROUP\n"
            'DEF_LOOKUP "SomeSub" SKIP_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "A"\n'
            'WITH GLYPH "A.c2sc"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        )
        self.assertEqual(
            fea,
            "# Glyph classes\n"
            "@SomeMarks = [marka markb];\n"
            "\n"
            "# Lookups\n"
            "lookup SomeSub {\n"
            "    lookupflag IgnoreBaseGlyphs;\n"
            "    sub A by A.c2sc;\n"
            "} SomeSub;\n",
        )

    def test_substitution_process_base(self):
        fea = self.parse(
            'DEF_GROUP "SomeMarks" ENUM GLYPH "marka" GLYPH "markb" '
            "END_ENUM END_GROUP\n"
            'DEF_LOOKUP "SomeSub" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "A"\n'
            'WITH GLYPH "A.c2sc"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        )
        self.assertEqual(
            fea,
            "# Glyph classes\n"
            "@SomeMarks = [marka markb];\n"
            "\n"
            "# Lookups\n"
            "lookup SomeSub {\n"
            "    sub A by A.c2sc;\n"
            "} SomeSub;\n",
        )

    def test_substitution_process_marks_all(self):
        fea = self.parse(
            'DEF_GROUP "SomeMarks" ENUM GLYPH "marka" GLYPH "markb" '
            "END_ENUM END_GROUP\n"
            'DEF_LOOKUP "SomeSub" PROCESS_BASE PROCESS_MARKS "ALL"'
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "A"\n'
            'WITH GLYPH "A.c2sc"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        )
        self.assertEqual(
            fea,
            "# Glyph classes\n"
            "@SomeMarks = [marka markb];\n"
            "\n"
            "# Lookups\n"
            "lookup SomeSub {\n"
            "    sub A by A.c2sc;\n"
            "} SomeSub;\n",
        )

    def test_substitution_process_marks_none(self):
        fea = self.parse(
            'DEF_GROUP "SomeMarks" ENUM GLYPH "marka" GLYPH "markb" '
            "END_ENUM END_GROUP\n"
            'DEF_LOOKUP "SomeSub" PROCESS_BASE PROCESS_MARKS "NONE"'
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "A"\n'
            'WITH GLYPH "A.c2sc"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        )
        self.assertEqual(
            fea,
            "# Glyph classes\n"
            "@SomeMarks = [marka markb];\n"
            "\n"
            "# Lookups\n"
            "lookup SomeSub {\n"
            "    lookupflag IgnoreMarks;\n"
            "    sub A by A.c2sc;\n"
            "} SomeSub;\n",
        )

    def test_substitution_skip_marks(self):
        fea = self.parse(
            'DEF_GROUP "SomeMarks" ENUM GLYPH "marka" GLYPH "markb" '
            "END_ENUM END_GROUP\n"
            'DEF_LOOKUP "SomeSub" PROCESS_BASE SKIP_MARKS '
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "A"\n'
            'WITH GLYPH "A.c2sc"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        )
        self.assertEqual(
            fea,
            "# Glyph classes\n"
            "@SomeMarks = [marka markb];\n"
            "\n"
            "# Lookups\n"
            "lookup SomeSub {\n"
            "    lookupflag IgnoreMarks;\n"
            "    sub A by A.c2sc;\n"
            "} SomeSub;\n",
        )

    def test_substitution_mark_attachment(self):
        fea = self.parse(
            'DEF_GROUP "SomeMarks" ENUM GLYPH "acutecmb" GLYPH "gravecmb" '
            "END_ENUM END_GROUP\n"
            'DEF_LOOKUP "SomeSub" PROCESS_BASE '
            'PROCESS_MARKS "SomeMarks" \n'
            "DIRECTION RTL\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "A"\n'
            'WITH GLYPH "A.c2sc"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        )
        self.assertEqual(
            fea,
            "# Glyph classes\n"
            "@SomeMarks = [acutecmb gravecmb];\n"
            "\n"
            "# Lookups\n"
            "lookup SomeSub {\n"
            "    lookupflag RightToLeft MarkAttachmentType"
            " @SomeMarks;\n"
            "    sub A by A.c2sc;\n"
            "} SomeSub;\n",
        )

    def test_substitution_mark_glyph_set(self):
        fea = self.parse(
            'DEF_GROUP "SomeMarks" ENUM GLYPH "acutecmb" GLYPH "gravecmb" '
            "END_ENUM END_GROUP\n"
            'DEF_LOOKUP "SomeSub" PROCESS_BASE '
            'PROCESS_MARKS MARK_GLYPH_SET "SomeMarks" \n'
            "DIRECTION RTL\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "A"\n'
            'WITH GLYPH "A.c2sc"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        )
        self.assertEqual(
            fea,
            "# Glyph classes\n"
            "@SomeMarks = [acutecmb gravecmb];\n"
            "\n"
            "# Lookups\n"
            "lookup SomeSub {\n"
            "    lookupflag RightToLeft UseMarkFilteringSet"
            " @SomeMarks;\n"
            "    sub A by A.c2sc;\n"
            "} SomeSub;\n",
        )

    def test_substitution_process_all_marks(self):
        fea = self.parse(
            'DEF_GROUP "SomeMarks" ENUM GLYPH "acutecmb" GLYPH "gravecmb" '
            "END_ENUM END_GROUP\n"
            'DEF_LOOKUP "SomeSub" PROCESS_BASE '
            "PROCESS_MARKS ALL \n"
            "DIRECTION RTL\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "A"\n'
            'WITH GLYPH "A.c2sc"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        )
        self.assertEqual(
            fea,
            "# Glyph classes\n"
            "@SomeMarks = [acutecmb gravecmb];\n"
            "\n"
            "# Lookups\n"
            "lookup SomeSub {\n"
            "    lookupflag RightToLeft;\n"
            "    sub A by A.c2sc;\n"
            "} SomeSub;\n",
        )

    def test_substitution_no_reversal(self):
        # TODO: check right context with no reversal
        fea = self.parse(
            'DEF_LOOKUP "Lookup" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            'RIGHT ENUM GLYPH "a" GLYPH "b" END_ENUM\n'
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "a"\n'
            'WITH GLYPH "a.alt"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        )
        self.assertEqual(
            fea,
            "\n# Lookups\n"
            "lookup Lookup {\n"
            "    sub a' [a b] by a.alt;\n"
            "} Lookup;\n",
        )

    def test_substitution_reversal(self):
        fea = self.parse(
            'DEF_GROUP "DFLT_Num_standardFigures"\n'
            'ENUM GLYPH "zero" GLYPH "one" GLYPH "two" END_ENUM\n'
            "END_GROUP\n"
            'DEF_GROUP "DFLT_Num_numerators"\n'
            'ENUM GLYPH "zero.numr" GLYPH "one.numr" GLYPH "two.numr" END_ENUM\n'
            "END_GROUP\n"
            'DEF_LOOKUP "RevLookup" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR REVERSAL\n"
            "IN_CONTEXT\n"
            'RIGHT ENUM GLYPH "a" GLYPH "b" END_ENUM\n'
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GROUP "DFLT_Num_standardFigures"\n'
            'WITH GROUP "DFLT_Num_numerators"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        )
        self.assertEqual(
            fea,
            "# Glyph classes\n"
            "@DFLT_Num_standardFigures = [zero one two];\n"
            "@DFLT_Num_numerators = [zero.numr one.numr two.numr];\n"
            "\n"
            "# Lookups\n"
            "lookup RevLookup {\n"
            "    rsub @DFLT_Num_standardFigures' [a b] by @DFLT_Num_numerators;\n"
            "} RevLookup;\n",
        )

    def test_substitution_single_to_multiple(self):
        fea = self.parse(
            'DEF_LOOKUP "ccmp" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "aacute"\n'
            'WITH GLYPH "a" GLYPH "acutecomb"\n'
            "END_SUB\n"
            'SUB GLYPH "agrave"\n'
            'WITH GLYPH "a" GLYPH "gravecomb"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        )
        self.assertEqual(
            fea,
            "\n# Lookups\n"
            "lookup ccmp {\n"
            "    sub aacute by a acutecomb;\n"
            "    sub agrave by a gravecomb;\n"
            "} ccmp;\n",
        )

    def test_substitution_multiple_to_single(self):
        fea = self.parse(
            'DEF_LOOKUP "liga" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB GLYPH "f" GLYPH "i"\n'
            'WITH GLYPH "f_i"\n'
            "END_SUB\n"
            'SUB GLYPH "f" GLYPH "t"\n'
            'WITH GLYPH "f_t"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        )
        self.assertEqual(
            fea,
            "\n# Lookups\n"
            "lookup liga {\n"
            "    sub f i by f_i;\n"
            "    sub f t by f_t;\n"
            "} liga;\n",
        )

    def test_substitution_reverse_chaining_single(self):
        fea = self.parse(
            'DEF_LOOKUP "numr" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR REVERSAL\n"
            "IN_CONTEXT\n"
            "RIGHT ENUM "
            'GLYPH "fraction" '
            'RANGE "zero.numr" TO "nine.numr" '
            "END_ENUM\n"
            "END_CONTEXT\n"
            "AS_SUBSTITUTION\n"
            'SUB RANGE "zero" TO "nine"\n'
            'WITH RANGE "zero.numr" TO "nine.numr"\n'
            "END_SUB\n"
            "END_SUBSTITUTION"
        )
        self.assertEqual(
            fea,
            "\n# Lookups\n"
            "lookup numr {\n"
            "    rsub zero - nine' [fraction zero.numr - nine.numr] by zero.numr - nine.numr;\n"
            "} numr;\n",
        )

    # GPOS
    #  ATTACH_CURSIVE
    #  ATTACH
    #  ADJUST_PAIR
    #  ADJUST_SINGLE
    def test_position_attach(self):
        fea = self.parse(
            'DEF_LOOKUP "anchor_top" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION RTL\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_POSITION\n"
            'ATTACH GLYPH "a" GLYPH "e"\n'
            'TO GLYPH "acutecomb" AT ANCHOR "top" '
            'GLYPH "gravecomb" AT ANCHOR "top"\n'
            "END_ATTACH\n"
            "END_POSITION\n"
            'DEF_ANCHOR "MARK_top" ON 120 GLYPH acutecomb COMPONENT 1 '
            "AT POS DX 0 DY 450 END_POS END_ANCHOR\n"
            'DEF_ANCHOR "MARK_top" ON 121 GLYPH gravecomb COMPONENT 1 '
            "AT POS DX 0 DY 450 END_POS END_ANCHOR\n"
            'DEF_ANCHOR "top" ON 31 GLYPH a COMPONENT 1 '
            "AT POS DX 210 DY 450 END_POS END_ANCHOR\n"
            'DEF_ANCHOR "top" ON 35 GLYPH e COMPONENT 1 '
            "AT POS DX 215 DY 450 END_POS END_ANCHOR\n"
        )
        self.assertEqual(
            fea,
            "\n# Mark classes\n"
            "markClass acutecomb <anchor 0 450> @top;\n"
            "markClass gravecomb <anchor 0 450> @top;\n"
            "\n"
            "# Lookups\n"
            "lookup anchor_top {\n"
            "    lookupflag RightToLeft;\n"
            "    pos base a\n"
            "        <anchor 210 450> mark @top;\n"
            "    pos base e\n"
            "        <anchor 215 450> mark @top;\n"
            "} anchor_top;\n",
        )

    def test_position_attach_mkmk(self):
        fea = self.parse(
            'DEF_GLYPH "brevecomb" ID 1 TYPE MARK END_GLYPH\n'
            'DEF_GLYPH "gravecomb" ID 2 TYPE MARK END_GLYPH\n'
            'DEF_LOOKUP "anchor_top" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION RTL\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_POSITION\n"
            'ATTACH GLYPH "gravecomb"\n'
            'TO GLYPH "acutecomb" AT ANCHOR "top"\n'
            "END_ATTACH\n"
            "END_POSITION\n"
            'DEF_ANCHOR "MARK_top" ON 1 GLYPH acutecomb COMPONENT 1 '
            "AT POS DX 0 DY 450 END_POS END_ANCHOR\n"
            'DEF_ANCHOR "top" ON 2 GLYPH gravecomb COMPONENT 1 '
            "AT POS DX 210 DY 450 END_POS END_ANCHOR\n"
        )
        self.assertEqual(
            fea,
            "\n# Mark classes\n"
            "markClass acutecomb <anchor 0 450> @top;\n"
            "\n"
            "# Lookups\n"
            "lookup anchor_top {\n"
            "    lookupflag RightToLeft;\n"
            "    pos mark gravecomb\n"
            "        <anchor 210 450> mark @top;\n"
            "} anchor_top;\n"
            "\n"
            "@GDEF_mark = [brevecomb gravecomb];\n"
            "table GDEF {\n"
            "    GlyphClassDef , , @GDEF_mark, ;\n"
            "} GDEF;\n",
        )

    def test_position_attach_in_context(self):
        fea = self.parse(
            'DEF_LOOKUP "test" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION RTL\n"
            'EXCEPT_CONTEXT LEFT GLYPH "a" END_CONTEXT\n'
            "AS_POSITION\n"
            'ATTACH GLYPH "a"\n'
            'TO GLYPH "acutecomb" AT ANCHOR "top" '
            'GLYPH "gravecomb" AT ANCHOR "top"\n'
            "END_ATTACH\n"
            "END_POSITION\n"
            'DEF_ANCHOR "MARK_top" ON 120 GLYPH acutecomb COMPONENT 1 '
            "AT POS DX 0 DY 450 END_POS END_ANCHOR\n"
            'DEF_ANCHOR "MARK_top" ON 121 GLYPH gravecomb COMPONENT 1 '
            "AT POS DX 0 DY 450 END_POS END_ANCHOR\n"
            'DEF_ANCHOR "top" ON 31 GLYPH a COMPONENT 1 '
            "AT POS DX 210 DY 450 END_POS END_ANCHOR\n"
        )
        self.assertEqual(
            fea,
            "\n# Mark classes\n"
            "markClass acutecomb <anchor 0 450> @top;\n"
            "markClass gravecomb <anchor 0 450> @top;\n"
            "\n"
            "# Lookups\n"
            "lookup test_target {\n"
            "    pos base a\n"
            "        <anchor 210 450> mark @top;\n"
            "} test_target;\n"
            "\n"
            "lookup test {\n"
            "    lookupflag RightToLeft;\n"
            "    ignore pos a [acutecomb gravecomb]';\n"
            "    pos [acutecomb gravecomb]' lookup test_target;\n"
            "} test;\n",
        )

    def test_position_attach_cursive(self):
        fea = self.parse(
            'DEF_LOOKUP "SomeLookup" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION RTL\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_POSITION\n"
            'ATTACH_CURSIVE EXIT GLYPH "a" GLYPH "b" '
            'ENTER GLYPH "a" GLYPH "c"\n'
            "END_ATTACH\n"
            "END_POSITION\n"
            'DEF_ANCHOR "exit"  ON 1 GLYPH a COMPONENT 1 AT POS END_POS END_ANCHOR\n'
            'DEF_ANCHOR "entry" ON 1 GLYPH a COMPONENT 1 AT POS END_POS END_ANCHOR\n'
            'DEF_ANCHOR "exit"  ON 2 GLYPH b COMPONENT 1 AT POS END_POS END_ANCHOR\n'
            'DEF_ANCHOR "entry" ON 3 GLYPH c COMPONENT 1 AT POS END_POS END_ANCHOR\n'
        )
        self.assertEqual(
            fea,
            "\n# Lookups\n"
            "lookup SomeLookup {\n"
            "    lookupflag RightToLeft;\n"
            "    pos cursive a <anchor 0 0> <anchor 0 0>;\n"
            "    pos cursive c <anchor 0 0> <anchor NULL>;\n"
            "    pos cursive b <anchor NULL> <anchor 0 0>;\n"
            "} SomeLookup;\n",
        )

    def test_position_adjust_pair(self):
        fea = self.parse(
            'DEF_LOOKUP "kern1" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION RTL\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_POSITION\n"
            "ADJUST_PAIR\n"
            ' FIRST GLYPH "A" FIRST GLYPH "V"\n'
            ' SECOND GLYPH "A" SECOND GLYPH "V"\n'
            " 1 2 BY POS ADV -30 END_POS POS END_POS\n"
            " 2 1 BY POS ADV -25 END_POS POS END_POS\n"
            "END_ADJUST\n"
            "END_POSITION\n"
        )
        self.assertEqual(
            fea,
            "\n# Lookups\n"
            "lookup kern1 {\n"
            "    lookupflag RightToLeft;\n"
            "    enum pos A V -30;\n"
            "    enum pos V A -25;\n"
            "} kern1;\n",
        )

    def test_position_adjust_pair_in_context(self):
        fea = self.parse(
            'DEF_LOOKUP "kern1" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            'EXCEPT_CONTEXT LEFT GLYPH "A" END_CONTEXT\n'
            "AS_POSITION\n"
            "ADJUST_PAIR\n"
            ' FIRST GLYPH "A" FIRST GLYPH "V"\n'
            ' SECOND GLYPH "A" SECOND GLYPH "V"\n'
            " 2 1 BY POS ADV -25 END_POS POS END_POS\n"
            "END_ADJUST\n"
            "END_POSITION\n"
        )
        self.assertEqual(
            fea,
            "\n# Lookups\n"
            "lookup kern1_target {\n"
            "    enum pos V A -25;\n"
            "} kern1_target;\n"
            "\n"
            "lookup kern1 {\n"
            "    ignore pos A V' A';\n"
            "    pos V' lookup kern1_target A' lookup kern1_target;\n"
            "} kern1;\n",
        )

    def test_position_adjust_single(self):
        fea = self.parse(
            'DEF_LOOKUP "TestLookup" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_POSITION\n"
            "ADJUST_SINGLE"
            ' GLYPH "glyph1" BY POS ADV 0 DX 123 END_POS\n'
            ' GLYPH "glyph2" BY POS ADV 0 DX 456 END_POS\n'
            "END_ADJUST\n"
            "END_POSITION\n"
        )
        self.assertEqual(
            fea,
            "\n# Lookups\n"
            "lookup TestLookup {\n"
            "    pos glyph1 <123 0 0 0>;\n"
            "    pos glyph2 <456 0 0 0>;\n"
            "} TestLookup;\n",
        )

    def test_position_adjust_single_in_context(self):
        fea = self.parse(
            'DEF_LOOKUP "TestLookup" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "EXCEPT_CONTEXT\n"
            'LEFT GLYPH "leftGlyph"\n'
            'RIGHT GLYPH "rightGlyph"\n'
            "END_CONTEXT\n"
            "AS_POSITION\n"
            "ADJUST_SINGLE"
            ' GLYPH "glyph1" BY POS ADV 0 DX 123 END_POS\n'
            ' GLYPH "glyph2" BY POS ADV 0 DX 456 END_POS\n'
            "END_ADJUST\n"
            "END_POSITION\n"
        )
        self.assertEqual(
            fea,
            "\n# Lookups\n"
            "lookup TestLookup_target {\n"
            "    pos glyph1 <123 0 0 0>;\n"
            "    pos glyph2 <456 0 0 0>;\n"
            "} TestLookup_target;\n"
            "\n"
            "lookup TestLookup {\n"
            "    ignore pos leftGlyph [glyph1 glyph2]' rightGlyph;\n"
            "    pos [glyph1 glyph2]' lookup TestLookup_target;\n"
            "} TestLookup;\n",
        )

    def test_def_anchor(self):
        fea = self.parse(
            'DEF_LOOKUP "TestLookup" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_POSITION\n"
            'ATTACH GLYPH "a"\n'
            'TO GLYPH "acutecomb" AT ANCHOR "top"\n'
            "END_ATTACH\n"
            "END_POSITION\n"
            'DEF_ANCHOR "top" ON 120 GLYPH a '
            "COMPONENT 1 AT POS DX 250 DY 450 END_POS END_ANCHOR\n"
            'DEF_ANCHOR "MARK_top" ON 120 GLYPH acutecomb '
            "COMPONENT 1 AT POS DX 0 DY 450 END_POS END_ANCHOR"
        )
        self.assertEqual(
            fea,
            "\n# Mark classes\n"
            "markClass acutecomb <anchor 0 450> @top;\n"
            "\n"
            "# Lookups\n"
            "lookup TestLookup {\n"
            "    pos base a\n"
            "        <anchor 250 450> mark @top;\n"
            "} TestLookup;\n",
        )

    def test_def_anchor_multi_component(self):
        fea = self.parse(
            'DEF_LOOKUP "TestLookup" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_POSITION\n"
            'ATTACH GLYPH "f_f"\n'
            'TO GLYPH "acutecomb" AT ANCHOR "top"\n'
            "END_ATTACH\n"
            "END_POSITION\n"
            'DEF_GLYPH "f_f" ID 120 TYPE LIGATURE COMPONENTS 2 END_GLYPH\n'
            'DEF_ANCHOR "top" ON 120 GLYPH f_f '
            "COMPONENT 1 AT POS DX 250 DY 450 END_POS END_ANCHOR\n"
            'DEF_ANCHOR "top" ON 120 GLYPH f_f '
            "COMPONENT 2 AT POS DX 450 DY 450 END_POS END_ANCHOR\n"
            'DEF_ANCHOR "MARK_top" ON 120 GLYPH acutecomb '
            "COMPONENT 1 AT POS  END_POS END_ANCHOR"
        )
        self.assertEqual(
            fea,
            "\n# Mark classes\n"
            "markClass acutecomb <anchor 0 0> @top;\n"
            "\n"
            "# Lookups\n"
            "lookup TestLookup {\n"
            "    pos ligature f_f\n"
            "            <anchor 250 450> mark @top\n"
            "        ligComponent\n"
            "            <anchor 450 450> mark @top;\n"
            "} TestLookup;\n"
            "\n"
            "@GDEF_ligature = [f_f];\n"
            "table GDEF {\n"
            "    GlyphClassDef , @GDEF_ligature, , ;\n"
            "} GDEF;\n",
        )

    def test_anchor_adjust_device(self):
        fea = self.parse(
            'DEF_ANCHOR "MARK_top" ON 123 GLYPH diacglyph '
            "COMPONENT 1 AT POS DX 0 DY 456 ADJUST_BY 12 AT 34 "
            "ADJUST_BY 56 AT 78 END_POS END_ANCHOR"
        )
        self.assertEqual(
            fea,
            "\n# Mark classes\n"
            "#markClass diacglyph <anchor 0 456 <device NULL>"
            " <device 34 12, 78 56>> @top;",
        )

    def test_use_extension(self):
        fea = self.parse(
            'DEF_LOOKUP "kern1" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR\n"
            "IN_CONTEXT\n"
            "END_CONTEXT\n"
            "AS_POSITION\n"
            "ADJUST_PAIR\n"
            ' FIRST GLYPH "A" FIRST GLYPH "V"\n'
            ' SECOND GLYPH "A" SECOND GLYPH "V"\n'
            " 1 2 BY POS ADV -30 END_POS POS END_POS\n"
            " 2 1 BY POS ADV -25 END_POS POS END_POS\n"
            "END_ADJUST\n"
            "END_POSITION\n"
            "COMPILER_USEEXTENSIONLOOKUPS\n"
        )
        self.assertEqual(
            fea,
            "\n# Lookups\n"
            "lookup kern1 useExtension {\n"
            "    enum pos A V -30;\n"
            "    enum pos V A -25;\n"
            "} kern1;\n",
        )

    def test_unsupported_compiler_flags(self):
        with self.assertLogs(level="WARNING") as logs:
            fea = self.parse("CMAP_FORMAT 0 3 4")
            self.assertEqual(fea, "")
        self.assertEqual(
            logs.output,
            [
                "WARNING:fontTools.voltLib.voltToFea:Unsupported setting ignored: CMAP_FORMAT"
            ],
        )

    def test_sanitize_lookup_name(self):
        fea = self.parse(
            'DEF_LOOKUP "Test Lookup" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR IN_CONTEXT END_CONTEXT\n"
            "AS_POSITION ADJUST_PAIR END_ADJUST END_POSITION\n"
            'DEF_LOOKUP "Test-Lookup" PROCESS_BASE PROCESS_MARKS ALL '
            "DIRECTION LTR IN_CONTEXT END_CONTEXT\n"
            "AS_POSITION ADJUST_PAIR END_ADJUST END_POSITION\n"
        )
        self.assertEqual(
            fea,
            "\n# Lookups\n"
            "lookup Test_Lookup {\n"
            "    \n"
            "} Test_Lookup;\n"
            "\n"
            "lookup Test_Lookup_ {\n"
            "    \n"
            "} Test_Lookup_;\n",
        )

    def test_sanitize_group_name(self):
        fea = self.parse(
            'DEF_GROUP "aaccented glyphs"\n'
            'ENUM GLYPH "aacute" GLYPH "abreve" END_ENUM\n'
            "END_GROUP\n"
            'DEF_GROUP "aaccented+glyphs"\n'
            'ENUM GLYPH "aacute" GLYPH "abreve" END_ENUM\n'
            "END_GROUP\n"
        )
        self.assertEqual(
            fea,
            "# Glyph classes\n"
            "@aaccented_glyphs = [aacute abreve];\n"
            "@aaccented_glyphs_ = [aacute abreve];",
        )

    def test_cli_vtp(self):
        vtp = DATADIR / "Nutso.vtp"
        fea = DATADIR / "Nutso.fea"
        self.cli(vtp, fea)

    def test_group_order(self):
        vtp = DATADIR / "NamdhinggoSIL1006.vtp"
        fea = DATADIR / "NamdhinggoSIL1006.fea"
        self.cli(vtp, fea)

    def test_cli_ttf(self):
        ttf = DATADIR / "Nutso.ttf"
        fea = DATADIR / "Nutso.fea"
        self.cli(ttf, fea)

    def test_cli_ttf_no_TSIV(self):
        from fontTools.voltLib.voltToFea import main as cli

        ttf = DATADIR / "Empty.ttf"
        temp = self.temp_path()
        self.assertEqual(1, cli([str(ttf), str(temp)]))

    def cli(self, source, fea):
        from fontTools.voltLib.voltToFea import main as cli

        temp = self.temp_path()
        cli([str(source), str(temp)])
        with temp.open() as f:
            res = f.read()
        with fea.open() as f:
            ref = f.read()
        self.assertEqual(ref, res)

    def parse(self, text):
        return VoltToFea(StringIO(text)).convert()


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



from fontTools.misc.transform import (
    Transform,
    Identity,
    Offset,
    Scale,
    DecomposedTransform,
)
import math
import pytest


class TransformTest(object):
    def test_examples(self):
        t = Transform()
        assert repr(t) == "<Transform [1 0 0 1 0 0]>"
        assert t.scale(2) == Transform(2, 0, 0, 2, 0, 0)
        assert t.scale(2.5, 5.5) == Transform(2.5, 0, 0, 5.5, 0, 0)
        assert t.scale(2, 3).transformPoint((100, 100)) == (200, 300)

    def test__init__(self):
        assert Transform(12) == Transform(12, 0, 0, 1, 0, 0)
        assert Transform(dx=12) == Transform(1, 0, 0, 1, 12, 0)
        assert Transform(yx=12) == Transform(1, 0, 12, 1, 0, 0)

    def test_transformPoints(self):
        t = Transform(2, 0, 0, 3, 0, 0)
        assert t.transformPoints([(0, 0), (0, 100), (100, 100), (100, 0)]) == [
            (0, 0),
            (0, 300),
            (200, 300),
            (200, 0),
        ]

    def test_transformVector(self):
        t = Transform(2, 0, 0, 3, -10, 30)
        assert t.transformVector((-4, 5)) == (-8, 15)

    def test_transformVectors(self):
        t = Transform(2, 0, 0, 3, -10, 30)
        assert t.transformVectors([(-4, 5), (-6, 7)]) == [(-8, 15), (-12, 21)]

    def test_translate(self):
        t = Transform()
        assert t.translate(20, 30) == Transform(1, 0, 0, 1, 20, 30)

    def test_scale(self):
        t = Transform()
        assert t.scale(5) == Transform(5, 0, 0, 5, 0, 0)
        assert t.scale(5, 6) == Transform(5, 0, 0, 6, 0, 0)

    def test_rotate(self):
        t = Transform()
        assert t.rotate(math.pi / 2) == Transform(0, 1, -1, 0, 0, 0)
        t = Transform()
        assert t.rotate(-math.pi / 2) == Transform(0, -1, 1, 0, 0, 0)
        t = Transform()
        assert tuple(t.rotate(math.radians(30))) == pytest.approx(
            tuple(Transform(0.866025, 0.5, -0.5, 0.866025, 0, 0))
        )

    def test_skew(self):
        t = Transform().skew(math.pi / 4)
        assert tuple(t) == pytest.approx(tuple(Transform(1, 0, 1, 1, 0, 0)))

    def test_transform(self):
        t = Transform(2, 0, 0, 3, 1, 6)
        assert t.transform((4, 3, 2, 1, 5, 6)) == Transform(8, 9, 4, 3, 11, 24)

    def test_reverseTransform(self):
        t = Transform(2, 0, 0, 3, 1, 6)
        reverse_t = t.reverseTransform((4, 3, 2, 1, 5, 6))
        assert reverse_t == Transform(8, 6, 6, 3, 21, 15)
        t = Transform(4, 3, 2, 1, 5, 6)
        reverse_t = t.transform((2, 0, 0, 3, 1, 6))
        assert reverse_t == Transform(8, 6, 6, 3, 21, 15)

    def test_inverse(self):
        t = Transform().translate(2, 3).scale(4, 5)
        assert t.transformPoint((10, 20)) == (42, 103)
        it = t.inverse()
        assert it.transformPoint((42, 103)) == (10.0, 20.0)
        assert Transform().inverse() == Transform()

    def test_toPS(self):
        t = Transform().scale(2, 3).translate(4, 5)
        assert t.toPS() == "[2 0 0 3 8 15]"

    def test__ne__(self):
        assert Transform() != Transform(2, 0, 0, 2, 0, 0)

    def test__hash__(self):
        t = Transform(12, 0, 0, 13, 0, 0)
        d = {t: None}
        assert t in d.keys()

    def test__bool__(self):
        assert not bool(Transform())
        assert Transform(2, 0, 0, 2, 0, 0)
        assert Transform(1, 0, 0, 1, 1, 0)

    def test__repr__(self):
        assert repr(Transform(1, 2, 3, 4, 5, 6)) == "<Transform [1 2 3 4 5 6]>"

    def test_Identity(self):
        assert isinstance(Identity, Transform)
        assert Identity == Transform(1, 0, 0, 1, 0, 0)

    def test_Offset(self):
        assert Offset() == Transform(1, 0, 0, 1, 0, 0)
        assert Offset(1) == Transform(1, 0, 0, 1, 1, 0)
        assert Offset(1, 2) == Transform(1, 0, 0, 1, 1, 2)

    def test_Scale(self):
        assert Scale(1) == Transform(1, 0, 0, 1, 0, 0)
        assert Scale(2) == Transform(2, 0, 0, 2, 0, 0)
        assert Scale(1, 2) == Transform(1, 0, 0, 2, 0, 0)

    def test_decompose(self):
        t = Transform(2, 0, 0, 3, 5, 7)
        d = t.toDecomposed()
        assert d.scaleX == 2
        assert d.scaleY == 3
        assert d.translateX == 5
        assert d.translateY == 7

        t = Transform(-1, 0, 0, 1, 0, 0)
        d = t.toDecomposed()
        assert d.scaleX == -1
        assert d.scaleY == 1
        assert d.rotation == 0

        t = Transform(1, 0, 0, -1, 0, 0)
        d = t.toDecomposed()
        assert d.scaleX == 1
        assert d.scaleY == -1
        assert d.rotation == 0


class DecomposedTransformTest(object):
    def test_identity(self):
        t = DecomposedTransform()
        assert (
            repr(t)
            == "DecomposedTransform(translateX=0, translateY=0, rotation=0, scaleX=1, scaleY=1, skewX=0, skewY=0, tCenterX=0, tCenterY=0)"
        )
        assert t == DecomposedTransform(scaleX=1.0)

    def test_scale(self):
        t = DecomposedTransform(scaleX=2, scaleY=3)
        assert t.scaleX == 2
        assert t.scaleY == 3

    def test_toTransform(self):
        t = DecomposedTransform(scaleX=2, scaleY=3)
        assert t.toTransform() == (2, 0, 0, 3, 0, 0)

    @pytest.mark.parametrize(
        "decomposed",
        [
            DecomposedTransform(scaleX=1, scaleY=0),
            DecomposedTransform(scaleX=0, scaleY=1),
            DecomposedTransform(scaleX=1, scaleY=0, rotation=30),
            DecomposedTransform(scaleX=0, scaleY=1, rotation=30),
            DecomposedTransform(scaleX=1, scaleY=1),
            DecomposedTransform(scaleX=-1, scaleY=1),
            DecomposedTransform(scaleX=1, scaleY=-1),
            DecomposedTransform(scaleX=-1, scaleY=-1),
            DecomposedTransform(rotation=90),
            DecomposedTransform(rotation=-90),
            DecomposedTransform(skewX=45),
            DecomposedTransform(skewY=45),
            DecomposedTransform(scaleX=-1, skewX=45),
            DecomposedTransform(scaleX=-1, skewY=45),
            DecomposedTransform(scaleY=-1, skewX=45),
            DecomposedTransform(scaleY=-1, skewY=45),
            DecomposedTransform(scaleX=-1, skewX=45, rotation=30),
            DecomposedTransform(scaleX=-1, skewY=45, rotation=30),
            DecomposedTransform(scaleY=-1, skewX=45, rotation=30),
            DecomposedTransform(scaleY=-1, skewY=45, rotation=30),
            DecomposedTransform(scaleX=-1, skewX=45, rotation=-30),
            DecomposedTransform(scaleX=-1, skewY=45, rotation=-30),
            DecomposedTransform(scaleY=-1, skewX=45, rotation=-30),
            DecomposedTransform(scaleY=-1, skewY=45, rotation=-30),
            DecomposedTransform(scaleX=-2, skewX=45, rotation=30),
            DecomposedTransform(scaleX=-2, skewY=45, rotation=30),
            DecomposedTransform(scaleY=-2, skewX=45, rotation=30),
            DecomposedTransform(scaleY=-2, skewY=45, rotation=30),
            DecomposedTransform(scaleX=-2, skewX=45, rotation=-30),
            DecomposedTransform(scaleX=-2, skewY=45, rotation=-30),
            DecomposedTransform(scaleY=-2, skewX=45, rotation=-30),
            DecomposedTransform(scaleY=-2, skewY=45, rotation=-30),
        ],
    )
    def test_roundtrip(lst, decomposed):
        assert decomposed.toTransform().toDecomposed().toTransform() == pytest.approx(
            tuple(decomposed.toTransform())
        ), decomposed



# coding: utf-8
from fontTools.misc import etree
from collections import OrderedDict
import io
import pytest


@pytest.mark.parametrize(
    "xml",
    [
        (
            "<root>"
            '<element key="value">text</element>'
            "<element>text</element>tail"
            "<empty-element/>"
            "</root>"
        ),
        (
            "<root>\n"
            '  <element key="value">text</element>\n'
            "  <element>text</element>tail\n"
            "  <empty-element/>\n"
            "</root>"
        ),
        (
            '<axis default="400" maximum="1000" minimum="1" name="weight" tag="wght">'
            '<labelname xml:lang="fa-IR">قطر</labelname>'
            "</axis>"
        ),
    ],
    ids=["simple_xml_no_indent", "simple_xml_indent", "xml_ns_attrib_utf_8"],
)
def test_roundtrip_string(xml):
    root = etree.fromstring(xml.encode("utf-8"))
    result = etree.tostring(root, encoding="utf-8").decode("utf-8")
    assert result == xml


def test_pretty_print():
    root = etree.Element("root")
    attrs = OrderedDict([("c", "2"), ("b", "1"), ("a", "0")])
    etree.SubElement(root, "element", attrs).text = "text"
    etree.SubElement(root, "element").text = "text"
    root.append(etree.Element("empty-element"))

    result = etree.tostring(root, encoding="unicode", pretty_print=True)

    assert result == (
        "<root>\n"
        '  <element c="2" b="1" a="0">text</element>\n'
        "  <element>text</element>\n"
        "  <empty-element/>\n"
        "</root>\n"
    )



"""Module for reading TFM (TeX Font Metrics) files.

The TFM format is described in the TFtoPL WEB source code, whose typeset form
can be found on `CTAN <http://mirrors.ctan.org/info/knuth-pdf/texware/tftopl.pdf>`_.

	>>> from fontTools.tfmLib import TFM
	>>> tfm = TFM("Tests/tfmLib/data/cmr10.tfm")
	>>>
	>>> # Accessing an attribute gets you metadata.
	>>> tfm.checksum
	1274110073
	>>> tfm.designsize
	10.0
	>>> tfm.codingscheme
	'TeX text'
	>>> tfm.family
	'CMR'
	>>> tfm.seven_bit_safe_flag
	False
	>>> tfm.face
	234
	>>> tfm.extraheader
	{}
	>>> tfm.fontdimens
	{'SLANT': 0.0, 'SPACE': 0.33333396911621094, 'STRETCH': 0.16666698455810547, 'SHRINK': 0.11111164093017578, 'XHEIGHT': 0.4305553436279297, 'QUAD': 1.0000028610229492, 'EXTRASPACE': 0.11111164093017578}
	>>> # Accessing a character gets you its metrics.
	>>> # “width” is always available, other metrics are available only when
	>>> # applicable. All values are relative to “designsize”.
	>>> tfm.chars[ord("g")]
	{'width': 0.5000019073486328, 'height': 0.4305553436279297, 'depth': 0.1944446563720703, 'italic': 0.013888359069824219}
	>>> # Kerning and ligature can be accessed as well.
	>>> tfm.kerning[ord("c")]
	{104: -0.02777862548828125, 107: -0.02777862548828125}
	>>> tfm.ligatures[ord("f")]
	{105: ('LIG', 12), 102: ('LIG', 11), 108: ('LIG', 13)}
"""

from types import SimpleNamespace

from fontTools.misc.sstruct import calcsize, unpack, unpack2

SIZES_FORMAT = """
    >
    lf: h    # length of the entire file, in words
    lh: h    # length of the header data, in words
    bc: h    # smallest character code in the font
    ec: h    # largest character code in the font
    nw: h    # number of words in the width table
    nh: h    # number of words in the height table
    nd: h    # number of words in the depth table
    ni: h    # number of words in the italic correction table
    nl: h    # number of words in the ligature/kern table
    nk: h    # number of words in the kern table
    ne: h    # number of words in the extensible character table
    np: h    # number of font parameter words
"""

SIZES_SIZE = calcsize(SIZES_FORMAT)

FIXED_FORMAT = "12.20F"

HEADER_FORMAT1 = f"""
    >
    checksum:            L
    designsize:          {FIXED_FORMAT}
"""

HEADER_FORMAT2 = f"""
    {HEADER_FORMAT1}
    codingscheme:        40p
"""

HEADER_FORMAT3 = f"""
    {HEADER_FORMAT2}
    family:              20p
"""

HEADER_FORMAT4 = f"""
    {HEADER_FORMAT3}
    seven_bit_safe_flag: ?
    ignored:             x
    ignored:             x
    face:                B
"""

HEADER_SIZE1 = calcsize(HEADER_FORMAT1)
HEADER_SIZE2 = calcsize(HEADER_FORMAT2)
HEADER_SIZE3 = calcsize(HEADER_FORMAT3)
HEADER_SIZE4 = calcsize(HEADER_FORMAT4)

LIG_KERN_COMMAND = """
    >
    skip_byte: B
    next_char: B
    op_byte: B
    remainder: B
"""

BASE_PARAMS = [
    "SLANT",
    "SPACE",
    "STRETCH",
    "SHRINK",
    "XHEIGHT",
    "QUAD",
    "EXTRASPACE",
]

MATHSY_PARAMS = [
    "NUM1",
    "NUM2",
    "NUM3",
    "DENOM1",
    "DENOM2",
    "SUP1",
    "SUP2",
    "SUP3",
    "SUB1",
    "SUB2",
    "SUPDROP",
    "SUBDROP",
    "DELIM1",
    "DELIM2",
    "AXISHEIGHT",
]

MATHEX_PARAMS = [
    "DEFAULTRULETHICKNESS",
    "BIGOPSPACING1",
    "BIGOPSPACING2",
    "BIGOPSPACING3",
    "BIGOPSPACING4",
    "BIGOPSPACING5",
]

VANILLA = 0
MATHSY = 1
MATHEX = 2

UNREACHABLE = 0
PASSTHROUGH = 1
ACCESSABLE = 2

NO_TAG = 0
LIG_TAG = 1
LIST_TAG = 2
EXT_TAG = 3

STOP_FLAG = 128
KERN_FLAG = 128


class TFMException(Exception):
    def __init__(self, message):
        super().__init__(message)


class TFM:
    def __init__(self, file):
        self._read(file)

    def __repr__(self):
        return (
            f"<TFM"
            f" for {self.family}"
            f" in {self.codingscheme}"
            f" at {self.designsize:g}pt>"
        )

    def _read(self, file):
        if hasattr(file, "read"):
            data = file.read()
        else:
            with open(file, "rb") as fp:
                data = fp.read()

        self._data = data

        if len(data) < SIZES_SIZE:
            raise TFMException("Too short input file")

        sizes = SimpleNamespace()
        unpack2(SIZES_FORMAT, data, sizes)

        # Do some file structure sanity checks.
        # TeX and TFtoPL do additional functional checks and might even correct
        # “errors” in the input file, but we instead try to output the file as
        # it is as long as it is parsable, even if the data make no sense.

        if sizes.lf < 0:
            raise TFMException("The file claims to have negative or zero length!")

        if len(data) < sizes.lf * 4:
            raise TFMException("The file has fewer bytes than it claims!")

        for name, length in vars(sizes).items():
            if length < 0:
                raise TFMException("The subfile size: '{name}' is negative!")

        if sizes.lh < 2:
            raise TFMException(f"The header length is only {sizes.lh}!")

        if sizes.bc > sizes.ec + 1 or sizes.ec > 255:
            raise TFMException(
                f"The character code range {sizes.bc}..{sizes.ec} is illegal!"
            )

        if sizes.nw == 0 or sizes.nh == 0 or sizes.nd == 0 or sizes.ni == 0:
            raise TFMException("Incomplete subfiles for character dimensions!")

        if sizes.ne > 256:
            raise TFMException(f"There are {ne} extensible recipes!")

        if sizes.lf != (
            6
            + sizes.lh
            + (sizes.ec - sizes.bc + 1)
            + sizes.nw
            + sizes.nh
            + sizes.nd
            + sizes.ni
            + sizes.nl
            + sizes.nk
            + sizes.ne
            + sizes.np
        ):
            raise TFMException("Subfile sizes don’t add up to the stated total")

        # Subfile offsets, used in the helper function below. These all are
        # 32-bit word offsets not 8-bit byte offsets.
        char_base = 6 + sizes.lh - sizes.bc
        width_base = char_base + sizes.ec + 1
        height_base = width_base + sizes.nw
        depth_base = height_base + sizes.nh
        italic_base = depth_base + sizes.nd
        lig_kern_base = italic_base + sizes.ni
        kern_base = lig_kern_base + sizes.nl
        exten_base = kern_base + sizes.nk
        param_base = exten_base + sizes.ne

        # Helper functions for accessing individual data. If this looks
        # nonidiomatic Python, I blame the effect of reading the literate WEB
        # documentation of TFtoPL.
        def char_info(c):
            return 4 * (char_base + c)

        def width_index(c):
            return data[char_info(c)]

        def noneexistent(c):
            return c < sizes.bc or c > sizes.ec or width_index(c) == 0

        def height_index(c):
            return data[char_info(c) + 1] // 16

        def depth_index(c):
            return data[char_info(c) + 1] % 16

        def italic_index(c):
            return data[char_info(c) + 2] // 4

        def tag(c):
            return data[char_info(c) + 2] % 4

        def remainder(c):
            return data[char_info(c) + 3]

        def width(c):
            r = 4 * (width_base + width_index(c))
            return read_fixed(r, "v")["v"]

        def height(c):
            r = 4 * (height_base + height_index(c))
            return read_fixed(r, "v")["v"]

        def depth(c):
            r = 4 * (depth_base + depth_index(c))
            return read_fixed(r, "v")["v"]

        def italic(c):
            r = 4 * (italic_base + italic_index(c))
            return read_fixed(r, "v")["v"]

        def exten(c):
            return 4 * (exten_base + remainder(c))

        def lig_step(i):
            return 4 * (lig_kern_base + i)

        def lig_kern_command(i):
            command = SimpleNamespace()
            unpack2(LIG_KERN_COMMAND, data[i:], command)
            return command

        def kern(i):
            r = 4 * (kern_base + i)
            return read_fixed(r, "v")["v"]

        def param(i):
            return 4 * (param_base + i)

        def read_fixed(index, key, obj=None):
            ret = unpack2(f">;{key}:{FIXED_FORMAT}", data[index:], obj)
            return ret[0]

        # Set all attributes to empty values regardless of the header size.
        unpack(HEADER_FORMAT4, [0] * HEADER_SIZE4, self)

        offset = 24
        length = sizes.lh * 4
        self.extraheader = {}
        if length >= HEADER_SIZE4:
            rest = unpack2(HEADER_FORMAT4, data[offset:], self)[1]
            if self.face < 18:
                s = self.face % 2
                b = self.face // 2
                self.face = "MBL"[b % 3] + "RI"[s] + "RCE"[b // 3]
            for i in range(sizes.lh - HEADER_SIZE4 // 4):
                rest = unpack2(f">;HEADER{i + 18}:l", rest, self.extraheader)[1]
        elif length >= HEADER_SIZE3:
            unpack2(HEADER_FORMAT3, data[offset:], self)
        elif length >= HEADER_SIZE2:
            unpack2(HEADER_FORMAT2, data[offset:], self)
        elif length >= HEADER_SIZE1:
            unpack2(HEADER_FORMAT1, data[offset:], self)

        self.fonttype = VANILLA
        scheme = self.codingscheme.upper()
        if scheme.startswith("TEX MATH SY"):
            self.fonttype = MATHSY
        elif scheme.startswith("TEX MATH EX"):
            self.fonttype = MATHEX

        self.fontdimens = {}
        for i in range(sizes.np):
            name = f"PARAMETER{i+1}"
            if i <= 6:
                name = BASE_PARAMS[i]
            elif self.fonttype == MATHSY and i <= 21:
                name = MATHSY_PARAMS[i - 7]
            elif self.fonttype == MATHEX and i <= 12:
                name = MATHEX_PARAMS[i - 7]
            read_fixed(param(i), name, self.fontdimens)

        lig_kern_map = {}
        self.right_boundary_char = None
        self.left_boundary_char = None
        if sizes.nl > 0:
            cmd = lig_kern_command(lig_step(0))
            if cmd.skip_byte == 255:
                self.right_boundary_char = cmd.next_char

            cmd = lig_kern_command(lig_step((sizes.nl - 1)))
            if cmd.skip_byte == 255:
                self.left_boundary_char = 256
                r = 256 * cmd.op_byte + cmd.remainder
                lig_kern_map[self.left_boundary_char] = r

        self.chars = {}
        for c in range(sizes.bc, sizes.ec + 1):
            if width_index(c) > 0:
                self.chars[c] = info = {}
                info["width"] = width(c)
                if height_index(c) > 0:
                    info["height"] = height(c)
                if depth_index(c) > 0:
                    info["depth"] = depth(c)
                if italic_index(c) > 0:
                    info["italic"] = italic(c)
                char_tag = tag(c)
                if char_tag == NO_TAG:
                    pass
                elif char_tag == LIG_TAG:
                    lig_kern_map[c] = remainder(c)
                elif char_tag == LIST_TAG:
                    info["nextlarger"] = remainder(c)
                elif char_tag == EXT_TAG:
                    info["varchar"] = varchar = {}
                    for i in range(4):
                        part = data[exten(c) + i]
                        if i == 3 or part > 0:
                            name = "rep"
                            if i == 0:
                                name = "top"
                            elif i == 1:
                                name = "mid"
                            elif i == 2:
                                name = "bot"
                            if noneexistent(part):
                                varchar[name] = c
                            else:
                                varchar[name] = part

        self.ligatures = {}
        self.kerning = {}
        for c, i in sorted(lig_kern_map.items()):
            cmd = lig_kern_command(lig_step(i))
            if cmd.skip_byte > STOP_FLAG:
                i = 256 * cmd.op_byte + cmd.remainder

            while i < sizes.nl:
                cmd = lig_kern_command(lig_step(i))
                if cmd.skip_byte > STOP_FLAG:
                    pass
                else:
                    if cmd.op_byte >= KERN_FLAG:
                        r = 256 * (cmd.op_byte - KERN_FLAG) + cmd.remainder
                        self.kerning.setdefault(c, {})[cmd.next_char] = kern(r)
                    else:
                        r = cmd.op_byte
                        if r == 4 or (r > 7 and r != 11):
                            # Ligature step with nonstandard code, we output
                            # the code verbatim.
                            lig = r
                        else:
                            lig = ""
                            if r % 4 > 1:
                                lig += "/"
                            lig += "LIG"
                            if r % 2 != 0:
                                lig += "/"
                            while r > 3:
                                lig += ">"
                                r -= 4
                        self.ligatures.setdefault(c, {})[cmd.next_char] = (
                            lig,
                            cmd.remainder,
                        )

                if cmd.skip_byte >= STOP_FLAG:
                    break
                i += cmd.skip_byte + 1


if __name__ == "__main__":
    import sys

    tfm = TFM(sys.argv[1])
    print(
        "\n".join(
            x
            for x in [
                f"tfm.checksum={tfm.checksum}",
                f"tfm.designsize={tfm.designsize}",
                f"tfm.codingscheme={tfm.codingscheme}",
                f"tfm.fonttype={tfm.fonttype}",
                f"tfm.family={tfm.family}",
                f"tfm.seven_bit_safe_flag={tfm.seven_bit_safe_flag}",
                f"tfm.face={tfm.face}",
                f"tfm.extraheader={tfm.extraheader}",
                f"tfm.fontdimens={tfm.fontdimens}",
                f"tfm.right_boundary_char={tfm.right_boundary_char}",
                f"tfm.left_boundary_char={tfm.left_boundary_char}",
                f"tfm.kerning={tfm.kerning}",
                f"tfm.ligatures={tfm.ligatures}",
                f"tfm.chars={tfm.chars}",
            ]
        )
    )
    print(tfm)



import math
import pytest
from fontTools.misc.arrayTools import Vector as ArrayVector
from fontTools.misc.vector import Vector


def test_Vector():
    v = Vector((100, 200))
    assert repr(v) == "Vector((100, 200))"
    assert v == Vector((100, 200))
    assert v == Vector([100, 200])
    assert v == (100, 200)
    assert (100, 200) == v
    assert v == [100, 200]
    assert [100, 200] == v
    assert v is Vector(v)
    assert v + 10 == (110, 210)
    assert 10 + v == (110, 210)
    assert v + Vector((1, 2)) == (101, 202)
    assert v - Vector((1, 2)) == (99, 198)
    assert v * 2 == (200, 400)
    assert 2 * v == (200, 400)
    assert v * 0.5 == (50, 100)
    assert v / 2 == (50, 100)
    assert 2 / v == (0.02, 0.01)
    v = Vector((3, 4))
    assert abs(v) == 5  # length
    assert v.length() == 5
    assert v.normalized() == Vector((0.6, 0.8))
    assert abs(Vector((1, 1, 1))) == math.sqrt(3)
    assert bool(Vector((0, 0, 1)))
    assert not bool(Vector((0, 0, 0)))
    v1 = Vector((2, 3))
    v2 = Vector((3, 4))
    assert v1.dot(v2) == 18
    v = Vector((2, 4))
    assert round(v / 3) == (1, 1)
    with pytest.raises(
        AttributeError,
        match="'Vector' object has no attribute 'newAttr'",
    ):
        v.newAttr = 12


def test_deprecated():
    with pytest.warns(
        DeprecationWarning,
        match="fontTools.misc.arrayTools.Vector has been deprecated",
    ):
        ArrayVector((1, 2))
    with pytest.warns(
        DeprecationWarning,
        match="the 'keep' argument has been deprecated",
    ):
        Vector((1, 2), keep=True)
    v = Vector((1, 2))
    with pytest.warns(
        DeprecationWarning,
        match="the 'toInt' method has been deprecated",
    ):
        v.toInt()
    with pytest.warns(
        DeprecationWarning,
        match="the 'values' attribute has been deprecated",
    ):
        v.values
    with pytest.raises(
        AttributeError,
        match="the 'values' attribute has been deprecated",
    ):
        v.values = [12, 23]



"""\
MS VOLT ``.vtp`` to AFDKO ``.fea`` OpenType Layout converter.

Usage
-----

To convert a VTP project file:


.. code-block:: sh

    $ fonttools voltLib.voltToFea input.vtp output.fea

It is also possible convert font files with `TSIV` table (as saved from Volt),
in this case the glyph names used in the Volt project will be mapped to the
actual glyph names in the font files when written to the feature file:

.. code-block:: sh

    $ fonttools voltLib.voltToFea input.ttf output.fea

The ``--quiet`` option can be used to suppress warnings.

The ``--traceback`` can be used to get Python traceback in case of exceptions,
instead of suppressing the traceback.


Limitations
-----------

* Not all VOLT features are supported, the script will error if it it
  encounters something it does not understand. Please report an issue if this
  happens.
* AFDKO feature file syntax for mark positioning is awkward and does not allow
  setting the mark coverage. It also defines mark anchors globally, as a result
  some mark positioning lookups might cover many marks than what was in the VOLT
  file. This should not be an issue in practice, but if it is then the only way
  is to modify the VOLT file or the generated feature file manually to use unique
  mark anchors for each lookup.
* VOLT allows subtable breaks in any lookup type, but AFDKO feature file
  implementations vary in their support; currently AFDKO’s makeOTF supports
  subtable breaks in pair positioning lookups only, while FontTools’ feaLib
  support it for most substitution lookups and only some positioning lookups.
"""

import logging
import re
from io import StringIO

from fontTools.feaLib import ast
from fontTools.ttLib import TTFont, TTLibError
from fontTools.voltLib import ast as VAst
from fontTools.voltLib.parser import Parser as VoltParser

log = logging.getLogger("fontTools.voltLib.voltToFea")

TABLES = ["GDEF", "GSUB", "GPOS"]


class MarkClassDefinition(ast.MarkClassDefinition):
    def asFea(self, indent=""):
        res = ""
        if not getattr(self, "used", False):
            res += "#"
        res += ast.MarkClassDefinition.asFea(self, indent)
        return res


# For sorting voltLib.ast.GlyphDefinition, see its use below.
class Group:
    def __init__(self, group):
        self.name = group.name.lower()
        self.groups = [
            x.group.lower() for x in group.enum.enum if isinstance(x, VAst.GroupName)
        ]

    def __lt__(self, other):
        if self.name in other.groups:
            return True
        if other.name in self.groups:
            return False
        if self.groups and not other.groups:
            return False
        if not self.groups and other.groups:
            return True


class VoltToFea:
    _NOT_LOOKUP_NAME_RE = re.compile(r"[^A-Za-z_0-9.]")
    _NOT_CLASS_NAME_RE = re.compile(r"[^A-Za-z_0-9.\-]")

    def __init__(self, file_or_path, font=None):
        self._file_or_path = file_or_path
        self._font = font

        self._glyph_map = {}
        self._glyph_order = None

        self._gdef = {}
        self._glyphclasses = {}
        self._features = {}
        self._lookups = {}

        self._marks = set()
        self._ligatures = {}

        self._markclasses = {}
        self._anchors = {}

        self._settings = {}

        self._lookup_names = {}
        self._class_names = {}

    def _lookupName(self, name):
        if name not in self._lookup_names:
            res = self._NOT_LOOKUP_NAME_RE.sub("_", name)
            while res in self._lookup_names.values():
                res += "_"
            self._lookup_names[name] = res
        return self._lookup_names[name]

    def _className(self, name):
        if name not in self._class_names:
            res = self._NOT_CLASS_NAME_RE.sub("_", name)
            while res in self._class_names.values():
                res += "_"
            self._class_names[name] = res
        return self._class_names[name]

    def _collectStatements(self, doc, tables):
        # Collect and sort group definitions first, to make sure a group
        # definition that references other groups comes after them since VOLT
        # does not enforce such ordering, and feature file require it.
        groups = [s for s in doc.statements if isinstance(s, VAst.GroupDefinition)]
        for statement in sorted(groups, key=lambda x: Group(x)):
            self._groupDefinition(statement)

        for statement in doc.statements:
            if isinstance(statement, VAst.GlyphDefinition):
                self._glyphDefinition(statement)
            elif isinstance(statement, VAst.AnchorDefinition):
                if "GPOS" in tables:
                    self._anchorDefinition(statement)
            elif isinstance(statement, VAst.SettingDefinition):
                self._settingDefinition(statement)
            elif isinstance(statement, VAst.GroupDefinition):
                pass  # Handled above
            elif isinstance(statement, VAst.ScriptDefinition):
                self._scriptDefinition(statement)
            elif not isinstance(statement, VAst.LookupDefinition):
                raise NotImplementedError(statement)

        # Lookup definitions need to be handled last as they reference glyph
        # and mark classes that might be defined after them.
        for statement in doc.statements:
            if isinstance(statement, VAst.LookupDefinition):
                if statement.pos and "GPOS" not in tables:
                    continue
                if statement.sub and "GSUB" not in tables:
                    continue
                self._lookupDefinition(statement)

    def _buildFeatureFile(self, tables):
        doc = ast.FeatureFile()
        statements = doc.statements

        if self._glyphclasses:
            statements.append(ast.Comment("# Glyph classes"))
            statements.extend(self._glyphclasses.values())

        if self._markclasses:
            statements.append(ast.Comment("\n# Mark classes"))
            statements.extend(c[1] for c in sorted(self._markclasses.items()))

        if self._lookups:
            statements.append(ast.Comment("\n# Lookups"))
            for lookup in self._lookups.values():
                statements.extend(getattr(lookup, "targets", []))
                statements.append(lookup)

        # Prune features
        features = self._features.copy()
        for ftag in features:
            scripts = features[ftag]
            for stag in scripts:
                langs = scripts[stag]
                for ltag in langs:
                    langs[ltag] = [l for l in langs[ltag] if l.lower() in self._lookups]
                scripts[stag] = {t: l for t, l in langs.items() if l}
            features[ftag] = {t: s for t, s in scripts.items() if s}
        features = {t: f for t, f in features.items() if f}

        if features:
            statements.append(ast.Comment("# Features"))
            for ftag, scripts in features.items():
                feature = ast.FeatureBlock(ftag)
                stags = sorted(scripts, key=lambda k: 0 if k == "DFLT" else 1)
                for stag in stags:
                    feature.statements.append(ast.ScriptStatement(stag))
                    ltags = sorted(scripts[stag], key=lambda k: 0 if k == "dflt" else 1)
                    for ltag in ltags:
                        include_default = True if ltag == "dflt" else False
                        feature.statements.append(
                            ast.LanguageStatement(ltag, include_default=include_default)
                        )
                        for name in scripts[stag][ltag]:
                            lookup = self._lookups[name.lower()]
                            lookupref = ast.LookupReferenceStatement(lookup)
                            feature.statements.append(lookupref)
                statements.append(feature)

        if self._gdef and "GDEF" in tables:
            classes = []
            for name in ("BASE", "MARK", "LIGATURE", "COMPONENT"):
                if name in self._gdef:
                    classname = "GDEF_" + name.lower()
                    glyphclass = ast.GlyphClassDefinition(classname, self._gdef[name])
                    statements.append(glyphclass)
                    classes.append(ast.GlyphClassName(glyphclass))
                else:
                    classes.append(None)

            gdef = ast.TableBlock("GDEF")
            gdef.statements.append(ast.GlyphClassDefStatement(*classes))
            statements.append(gdef)

        return doc

    def convert(self, tables=None):
        doc = VoltParser(self._file_or_path).parse()

        if tables is None:
            tables = TABLES
        if self._font is not None:
            self._glyph_order = self._font.getGlyphOrder()

        self._collectStatements(doc, tables)
        fea = self._buildFeatureFile(tables)
        return fea.asFea()

    def _glyphName(self, glyph):
        try:
            name = glyph.glyph
        except AttributeError:
            name = glyph
        return ast.GlyphName(self._glyph_map.get(name, name))

    def _groupName(self, group):
        try:
            name = group.group
        except AttributeError:
            name = group
        return ast.GlyphClassName(self._glyphclasses[name.lower()])

    def _coverage(self, coverage):
        items = []
        for item in coverage:
            if isinstance(item, VAst.GlyphName):
                items.append(self._glyphName(item))
            elif isinstance(item, VAst.GroupName):
                items.append(self._groupName(item))
            elif isinstance(item, VAst.Enum):
                items.append(self._enum(item))
            elif isinstance(item, VAst.Range):
                items.append((item.start, item.end))
            else:
                raise NotImplementedError(item)
        return items

    def _enum(self, enum):
        return ast.GlyphClass(self._coverage(enum.enum))

    def _context(self, context):
        out = []
        for item in context:
            coverage = self._coverage(item)
            if not isinstance(coverage, (tuple, list)):
                coverage = [coverage]
            out.extend(coverage)
        return out

    def _groupDefinition(self, group):
        name = self._className(group.name)
        glyphs = self._enum(group.enum)
        glyphclass = ast.GlyphClassDefinition(name, glyphs)

        self._glyphclasses[group.name.lower()] = glyphclass

    def _glyphDefinition(self, glyph):
        try:
            self._glyph_map[glyph.name] = self._glyph_order[glyph.id]
        except TypeError:
            pass

        if glyph.type in ("BASE", "MARK", "LIGATURE", "COMPONENT"):
            if glyph.type not in self._gdef:
                self._gdef[glyph.type] = ast.GlyphClass()
            self._gdef[glyph.type].glyphs.append(self._glyphName(glyph.name))

        if glyph.type == "MARK":
            self._marks.add(glyph.name)
        elif glyph.type == "LIGATURE":
            self._ligatures[glyph.name] = glyph.components

    def _scriptDefinition(self, script):
        stag = script.tag
        for lang in script.langs:
            ltag = lang.tag
            for feature in lang.features:
                lookups = {l.split("\\")[0]: True for l in feature.lookups}
                ftag = feature.tag
                if ftag not in self._features:
                    self._features[ftag] = {}
                if stag not in self._features[ftag]:
                    self._features[ftag][stag] = {}
                assert ltag not in self._features[ftag][stag]
                self._features[ftag][stag][ltag] = lookups.keys()

    def _settingDefinition(self, setting):
        if setting.name.startswith("COMPILER_"):
            self._settings[setting.name] = setting.value
        else:
            log.warning(f"Unsupported setting ignored: {setting.name}")

    def _adjustment(self, adjustment):
        adv, dx, dy, adv_adjust_by, dx_adjust_by, dy_adjust_by = adjustment

        adv_device = adv_adjust_by and adv_adjust_by.items() or None
        dx_device = dx_adjust_by and dx_adjust_by.items() or None
        dy_device = dy_adjust_by and dy_adjust_by.items() or None

        return ast.ValueRecord(
            xPlacement=dx,
            yPlacement=dy,
            xAdvance=adv,
            xPlaDevice=dx_device,
            yPlaDevice=dy_device,
            xAdvDevice=adv_device,
        )

    def _anchor(self, adjustment):
        adv, dx, dy, adv_adjust_by, dx_adjust_by, dy_adjust_by = adjustment

        assert not adv_adjust_by
        dx_device = dx_adjust_by and dx_adjust_by.items() or None
        dy_device = dy_adjust_by and dy_adjust_by.items() or None

        return ast.Anchor(
            dx or 0,
            dy or 0,
            xDeviceTable=dx_device or None,
            yDeviceTable=dy_device or None,
        )

    def _anchorDefinition(self, anchordef):
        anchorname = anchordef.name
        glyphname = anchordef.glyph_name
        anchor = self._anchor(anchordef.pos)

        if anchorname.startswith("MARK_"):
            name = "_".join(anchorname.split("_")[1:])
            markclass = ast.MarkClass(self._className(name))
            glyph = self._glyphName(glyphname)
            markdef = MarkClassDefinition(markclass, anchor, glyph)
            self._markclasses[(glyphname, anchorname)] = markdef
        else:
            if glyphname not in self._anchors:
                self._anchors[glyphname] = {}
            if anchorname not in self._anchors[glyphname]:
                self._anchors[glyphname][anchorname] = {}
            self._anchors[glyphname][anchorname][anchordef.component] = anchor

    def _gposLookup(self, lookup, fealookup):
        statements = fealookup.statements

        pos = lookup.pos
        if isinstance(pos, VAst.PositionAdjustPairDefinition):
            for (idx1, idx2), (pos1, pos2) in pos.adjust_pair.items():
                coverage_1 = pos.coverages_1[idx1 - 1]
                coverage_2 = pos.coverages_2[idx2 - 1]

                # If not both are groups, use “enum pos” otherwise makeotf will
                # fail.
                enumerated = False
                for item in coverage_1 + coverage_2:
                    if not isinstance(item, VAst.GroupName):
                        enumerated = True

                glyphs1 = self._coverage(coverage_1)
                glyphs2 = self._coverage(coverage_2)
                record1 = self._adjustment(pos1)
                record2 = self._adjustment(pos2)
                assert len(glyphs1) == 1
                assert len(glyphs2) == 1
                statements.append(
                    ast.PairPosStatement(
                        glyphs1[0], record1, glyphs2[0], record2, enumerated=enumerated
                    )
                )
        elif isinstance(pos, VAst.PositionAdjustSingleDefinition):
            for a, b in pos.adjust_single:
                glyphs = self._coverage(a)
                record = self._adjustment(b)
                assert len(glyphs) == 1
                statements.append(
                    ast.SinglePosStatement([(glyphs[0], record)], [], [], False)
                )
        elif isinstance(pos, VAst.PositionAttachDefinition):
            anchors = {}
            for marks, classname in pos.coverage_to:
                for mark in marks:
                    # Set actually used mark classes. Basically a hack to get
                    # around the feature file syntax limitation of making mark
                    # classes global and not allowing mark positioning to
                    # specify mark coverage.
                    for name in mark.glyphSet():
                        key = (name, "MARK_" + classname)
                        self._markclasses[key].used = True
                markclass = ast.MarkClass(self._className(classname))
                for base in pos.coverage:
                    for name in base.glyphSet():
                        if name not in anchors:
                            anchors[name] = []
                        if classname not in anchors[name]:
                            anchors[name].append(classname)

            for name in anchors:
                components = 1
                if name in self._ligatures:
                    components = self._ligatures[name]

                marks = []
                for mark in anchors[name]:
                    markclass = ast.MarkClass(self._className(mark))
                    for component in range(1, components + 1):
                        if len(marks) < component:
                            marks.append([])
                        anchor = None
                        if component in self._anchors[name][mark]:
                            anchor = self._anchors[name][mark][component]
                        marks[component - 1].append((anchor, markclass))

                base = self._glyphName(name)
                if name in self._marks:
                    mark = ast.MarkMarkPosStatement(base, marks[0])
                elif name in self._ligatures:
                    mark = ast.MarkLigPosStatement(base, marks)
                else:
                    mark = ast.MarkBasePosStatement(base, marks[0])
                statements.append(mark)
        elif isinstance(pos, VAst.PositionAttachCursiveDefinition):
            # Collect enter and exit glyphs
            enter_coverage = []
            for coverage in pos.coverages_enter:
                for base in coverage:
                    for name in base.glyphSet():
                        enter_coverage.append(name)
            exit_coverage = []
            for coverage in pos.coverages_exit:
                for base in coverage:
                    for name in base.glyphSet():
                        exit_coverage.append(name)

            # Write enter anchors, also check if the glyph has exit anchor and
            # write it, too.
            for name in enter_coverage:
                glyph = self._glyphName(name)
                entry = self._anchors[name]["entry"][1]
                exit = None
                if name in exit_coverage:
                    exit = self._anchors[name]["exit"][1]
                    exit_coverage.pop(exit_coverage.index(name))
                statements.append(ast.CursivePosStatement(glyph, entry, exit))

            # Write any remaining exit anchors.
            for name in exit_coverage:
                glyph = self._glyphName(name)
                exit = self._anchors[name]["exit"][1]
                statements.append(ast.CursivePosStatement(glyph, None, exit))
        else:
            raise NotImplementedError(pos)

    def _gposContextLookup(
        self, lookup, prefix, suffix, ignore, fealookup, targetlookup
    ):
        statements = fealookup.statements

        assert not lookup.reversal

        pos = lookup.pos
        if isinstance(pos, VAst.PositionAdjustPairDefinition):
            for (idx1, idx2), (pos1, pos2) in pos.adjust_pair.items():
                glyphs1 = self._coverage(pos.coverages_1[idx1 - 1])
                glyphs2 = self._coverage(pos.coverages_2[idx2 - 1])
                assert len(glyphs1) == 1
                assert len(glyphs2) == 1
                glyphs = (glyphs1[0], glyphs2[0])

                if ignore:
                    statement = ast.IgnorePosStatement([(prefix, glyphs, suffix)])
                else:
                    lookups = (targetlookup, targetlookup)
                    statement = ast.ChainContextPosStatement(
                        prefix, glyphs, suffix, lookups
                    )
                statements.append(statement)
        elif isinstance(pos, VAst.PositionAdjustSingleDefinition):
            glyphs = [ast.GlyphClass()]
            for a, b in pos.adjust_single:
                glyph = self._coverage(a)
                glyphs[0].extend(glyph)

            if ignore:
                statement = ast.IgnorePosStatement([(prefix, glyphs, suffix)])
            else:
                statement = ast.ChainContextPosStatement(
                    prefix, glyphs, suffix, [targetlookup]
                )
            statements.append(statement)
        elif isinstance(pos, VAst.PositionAttachDefinition):
            glyphs = [ast.GlyphClass()]
            for coverage, _ in pos.coverage_to:
                glyphs[0].extend(self._coverage(coverage))

            if ignore:
                statement = ast.IgnorePosStatement([(prefix, glyphs, suffix)])
            else:
                statement = ast.ChainContextPosStatement(
                    prefix, glyphs, suffix, [targetlookup]
                )
            statements.append(statement)
        else:
            raise NotImplementedError(pos)

    def _gsubLookup(self, lookup, prefix, suffix, ignore, chain, fealookup):
        statements = fealookup.statements

        sub = lookup.sub
        for key, val in sub.mapping.items():
            if not key or not val:
                path, line, column = sub.location
                log.warning(f"{path}:{line}:{column}: Ignoring empty substitution")
                continue
            statement = None
            glyphs = self._coverage(key)
            replacements = self._coverage(val)
            if ignore:
                chain_context = (prefix, glyphs, suffix)
                statement = ast.IgnoreSubstStatement([chain_context])
            elif isinstance(sub, VAst.SubstitutionSingleDefinition):
                assert len(glyphs) == 1
                assert len(replacements) == 1
                statement = ast.SingleSubstStatement(
                    glyphs, replacements, prefix, suffix, chain
                )
            elif isinstance(sub, VAst.SubstitutionReverseChainingSingleDefinition):
                assert len(glyphs) == 1
                assert len(replacements) == 1
                statement = ast.ReverseChainSingleSubstStatement(
                    prefix, suffix, glyphs, replacements
                )
            elif isinstance(sub, VAst.SubstitutionMultipleDefinition):
                assert len(glyphs) == 1
                statement = ast.MultipleSubstStatement(
                    prefix, glyphs[0], suffix, replacements, chain
                )
            elif isinstance(sub, VAst.SubstitutionLigatureDefinition):
                assert len(replacements) == 1
                statement = ast.LigatureSubstStatement(
                    prefix, glyphs, suffix, replacements[0], chain
                )
            else:
                raise NotImplementedError(sub)
            statements.append(statement)

    def _lookupDefinition(self, lookup):
        mark_attachement = None
        mark_filtering = None

        flags = 0
        if lookup.direction == "RTL":
            flags |= 1
        if not lookup.process_base:
            flags |= 2
        # FIXME: Does VOLT support this?
        # if not lookup.process_ligatures:
        #     flags |= 4
        if not lookup.process_marks:
            flags |= 8
        elif isinstance(lookup.process_marks, str):
            mark_attachement = self._groupName(lookup.process_marks)
        elif lookup.mark_glyph_set is not None:
            mark_filtering = self._groupName(lookup.mark_glyph_set)

        lookupflags = None
        if flags or mark_attachement is not None or mark_filtering is not None:
            lookupflags = ast.LookupFlagStatement(
                flags, mark_attachement, mark_filtering
            )
        if "\\" in lookup.name:
            # Merge sub lookups as subtables (lookups named “base\sub”),
            # makeotf/feaLib will issue a warning and ignore the subtable
            # statement if it is not a pairpos lookup, though.
            name = lookup.name.split("\\")[0]
            if name.lower() not in self._lookups:
                fealookup = ast.LookupBlock(self._lookupName(name))
                if lookupflags is not None:
                    fealookup.statements.append(lookupflags)
                fealookup.statements.append(ast.Comment("# " + lookup.name))
            else:
                fealookup = self._lookups[name.lower()]
                fealookup.statements.append(ast.SubtableStatement())
                fealookup.statements.append(ast.Comment("# " + lookup.name))
            self._lookups[name.lower()] = fealookup
        else:
            fealookup = ast.LookupBlock(self._lookupName(lookup.name))
            if lookupflags is not None:
                fealookup.statements.append(lookupflags)
            self._lookups[lookup.name.lower()] = fealookup

        if lookup.comments is not None:
            fealookup.statements.append(ast.Comment("# " + lookup.comments))

        contexts = []
        if lookup.context:
            for context in lookup.context:
                prefix = self._context(context.left)
                suffix = self._context(context.right)
                ignore = context.ex_or_in == "EXCEPT_CONTEXT"
                contexts.append([prefix, suffix, ignore, False])
                # It seems that VOLT will create contextual substitution using
                # only the input if there is no other contexts in this lookup.
                if ignore and len(lookup.context) == 1:
                    contexts.append([[], [], False, True])
        else:
            contexts.append([[], [], False, False])

        targetlookup = None
        for prefix, suffix, ignore, chain in contexts:
            if lookup.sub is not None:
                self._gsubLookup(lookup, prefix, suffix, ignore, chain, fealookup)

            if lookup.pos is not None:
                if self._settings.get("COMPILER_USEEXTENSIONLOOKUPS"):
                    fealookup.use_extension = True
                if prefix or suffix or chain or ignore:
                    if not ignore and targetlookup is None:
                        targetname = self._lookupName(lookup.name + " target")
                        targetlookup = ast.LookupBlock(targetname)
                        fealookup.targets = getattr(fealookup, "targets", [])
                        fealookup.targets.append(targetlookup)
                        self._gposLookup(lookup, targetlookup)
                    self._gposContextLookup(
                        lookup, prefix, suffix, ignore, fealookup, targetlookup
                    )
                else:
                    self._gposLookup(lookup, fealookup)


def main(args=None):
    """Convert MS VOLT to AFDKO feature files."""

    import argparse
    from pathlib import Path

    from fontTools import configLogger

    parser = argparse.ArgumentParser(
        "fonttools voltLib.voltToFea", description=main.__doc__
    )
    parser.add_argument(
        "input", metavar="INPUT", type=Path, help="input font/VTP file to process"
    )
    parser.add_argument(
        "featurefile", metavar="OUTPUT", type=Path, help="output feature file"
    )
    parser.add_argument(
        "-t",
        "--table",
        action="append",
        choices=TABLES,
        dest="tables",
        help="List of tables to write, by default all tables are written",
    )
    parser.add_argument(
        "-q", "--quiet", action="store_true", help="Suppress non-error messages"
    )
    parser.add_argument(
        "--traceback", action="store_true", help="Don’t catch exceptions"
    )

    options = parser.parse_args(args)

    configLogger(level=("ERROR" if options.quiet else "INFO"))

    file_or_path = options.input
    font = None
    try:
        font = TTFont(file_or_path)
        if "TSIV" in font:
            file_or_path = StringIO(font["TSIV"].data.decode("utf-8"))
        else:
            log.error('"TSIV" table is missing, font was not saved from VOLT?')
            return 1
    except TTLibError:
        pass

    converter = VoltToFea(file_or_path, font)
    try:
        fea = converter.convert(options.tables)
    except NotImplementedError as e:
        if options.traceback:
            raise
        location = getattr(e.args[0], "location", None)
        message = f'"{e}" is not supported'
        if location:
            path, line, column = location
            log.error(f"{path}:{line}:{column}: {message}")
        else:
            log.error(message)
        return 1
    with open(options.featurefile, "w") as feafile:
        feafile.write(fea)


if __name__ == "__main__":
    import sys

    sys.exit(main())



from fontTools.misc.psCharStrings import (
    SimpleT2Decompiler,
    T2WidthExtractor,
    calcSubrBias,
)


def _uniq_sort(l):
    return sorted(set(l))


class StopHintCountEvent(Exception):
    pass


class _DesubroutinizingT2Decompiler(SimpleT2Decompiler):
    stop_hintcount_ops = (
        "op_hintmask",
        "op_cntrmask",
        "op_rmoveto",
        "op_hmoveto",
        "op_vmoveto",
    )

    def __init__(self, localSubrs, globalSubrs, private=None):
        SimpleT2Decompiler.__init__(self, localSubrs, globalSubrs, private)

    def execute(self, charString):
        self.need_hintcount = True  # until proven otherwise
        for op_name in self.stop_hintcount_ops:
            setattr(self, op_name, self.stop_hint_count)

        if hasattr(charString, "_desubroutinized"):
            # If a charstring has already been desubroutinized, we will still
            # need to execute it if we need to count hints in order to
            # compute the byte length for mask arguments, and haven't finished
            # counting hints pairs.
            if self.need_hintcount and self.callingStack:
                try:
                    SimpleT2Decompiler.execute(self, charString)
                except StopHintCountEvent:
                    del self.callingStack[-1]
            return

        charString._patches = []
        SimpleT2Decompiler.execute(self, charString)
        desubroutinized = charString.program[:]
        for idx, expansion in reversed(charString._patches):
            assert idx >= 2
            assert desubroutinized[idx - 1] in [
                "callsubr",
                "callgsubr",
            ], desubroutinized[idx - 1]
            assert type(desubroutinized[idx - 2]) == int
            if expansion[-1] == "return":
                expansion = expansion[:-1]
            desubroutinized[idx - 2 : idx] = expansion
        if not self.private.in_cff2:
            if "endchar" in desubroutinized:
                # Cut off after first endchar
                desubroutinized = desubroutinized[
                    : desubroutinized.index("endchar") + 1
                ]

        charString._desubroutinized = desubroutinized
        del charString._patches

    def op_callsubr(self, index):
        subr = self.localSubrs[self.operandStack[-1] + self.localBias]
        SimpleT2Decompiler.op_callsubr(self, index)
        self.processSubr(index, subr)

    def op_callgsubr(self, index):
        subr = self.globalSubrs[self.operandStack[-1] + self.globalBias]
        SimpleT2Decompiler.op_callgsubr(self, index)
        self.processSubr(index, subr)

    def stop_hint_count(self, *args):
        self.need_hintcount = False
        for op_name in self.stop_hintcount_ops:
            setattr(self, op_name, None)
        cs = self.callingStack[-1]
        if hasattr(cs, "_desubroutinized"):
            raise StopHintCountEvent()

    def op_hintmask(self, index):
        SimpleT2Decompiler.op_hintmask(self, index)
        if self.need_hintcount:
            self.stop_hint_count()

    def processSubr(self, index, subr):
        cs = self.callingStack[-1]
        if not hasattr(cs, "_desubroutinized"):
            cs._patches.append((index, subr._desubroutinized))


def desubroutinize(cff):
    for fontName in cff.fontNames:
        font = cff[fontName]
        cs = font.CharStrings
        for c in cs.values():
            c.decompile()
            subrs = getattr(c.private, "Subrs", [])
            decompiler = _DesubroutinizingT2Decompiler(subrs, c.globalSubrs, c.private)
            decompiler.execute(c)
            c.program = c._desubroutinized
            del c._desubroutinized
        # Delete all the local subrs
        if hasattr(font, "FDArray"):
            for fd in font.FDArray:
                pd = fd.Private
                if hasattr(pd, "Subrs"):
                    del pd.Subrs
                if "Subrs" in pd.rawDict:
                    del pd.rawDict["Subrs"]
        else:
            pd = font.Private
            if hasattr(pd, "Subrs"):
                del pd.Subrs
            if "Subrs" in pd.rawDict:
                del pd.rawDict["Subrs"]
    # as well as the global subrs
    cff.GlobalSubrs.clear()


class _MarkingT2Decompiler(SimpleT2Decompiler):
    def __init__(self, localSubrs, globalSubrs, private):
        SimpleT2Decompiler.__init__(self, localSubrs, globalSubrs, private)
        for subrs in [localSubrs, globalSubrs]:
            if subrs and not hasattr(subrs, "_used"):
                subrs._used = set()

    def op_callsubr(self, index):
        self.localSubrs._used.add(self.operandStack[-1] + self.localBias)
        SimpleT2Decompiler.op_callsubr(self, index)

    def op_callgsubr(self, index):
        self.globalSubrs._used.add(self.operandStack[-1] + self.globalBias)
        SimpleT2Decompiler.op_callgsubr(self, index)


class _DehintingT2Decompiler(T2WidthExtractor):
    class Hints(object):
        def __init__(self):
            # Whether calling this charstring produces any hint stems
            # Note that if a charstring starts with hintmask, it will
            # have has_hint set to True, because it *might* produce an
            # implicit vstem if called under certain conditions.
            self.has_hint = False
            # Index to start at to drop all hints
            self.last_hint = 0
            # Index up to which we know more hints are possible.
            # Only relevant if status is 0 or 1.
            self.last_checked = 0
            # The status means:
            # 0: after dropping hints, this charstring is empty
            # 1: after dropping hints, there may be more hints
            # 	continuing after this, or there might be
            # 	other things.  Not clear yet.
            # 2: no more hints possible after this charstring
            self.status = 0
            # Has hintmask instructions; not recursive
            self.has_hintmask = False
            # List of indices of calls to empty subroutines to remove.
            self.deletions = []

        pass

    def __init__(
        self, css, localSubrs, globalSubrs, nominalWidthX, defaultWidthX, private=None
    ):
        self._css = css
        T2WidthExtractor.__init__(
            self, localSubrs, globalSubrs, nominalWidthX, defaultWidthX
        )
        self.private = private

    def execute(self, charString):
        old_hints = charString._hints if hasattr(charString, "_hints") else None
        charString._hints = self.Hints()

        T2WidthExtractor.execute(self, charString)

        hints = charString._hints

        if hints.has_hint or hints.has_hintmask:
            self._css.add(charString)

        if hints.status != 2:
            # Check from last_check, make sure we didn't have any operators.
            for i in range(hints.last_checked, len(charString.program) - 1):
                if isinstance(charString.program[i], str):
                    hints.status = 2
                    break
                else:
                    hints.status = 1  # There's *something* here
            hints.last_checked = len(charString.program)

        if old_hints:
            assert hints.__dict__ == old_hints.__dict__

    def op_callsubr(self, index):
        subr = self.localSubrs[self.operandStack[-1] + self.localBias]
        T2WidthExtractor.op_callsubr(self, index)
        self.processSubr(index, subr)

    def op_callgsubr(self, index):
        subr = self.globalSubrs[self.operandStack[-1] + self.globalBias]
        T2WidthExtractor.op_callgsubr(self, index)
        self.processSubr(index, subr)

    def op_hstem(self, index):
        T2WidthExtractor.op_hstem(self, index)
        self.processHint(index)

    def op_vstem(self, index):
        T2WidthExtractor.op_vstem(self, index)
        self.processHint(index)

    def op_hstemhm(self, index):
        T2WidthExtractor.op_hstemhm(self, index)
        self.processHint(index)

    def op_vstemhm(self, index):
        T2WidthExtractor.op_vstemhm(self, index)
        self.processHint(index)

    def op_hintmask(self, index):
        rv = T2WidthExtractor.op_hintmask(self, index)
        self.processHintmask(index)
        return rv

    def op_cntrmask(self, index):
        rv = T2WidthExtractor.op_cntrmask(self, index)
        self.processHintmask(index)
        return rv

    def processHintmask(self, index):
        cs = self.callingStack[-1]
        hints = cs._hints
        hints.has_hintmask = True
        if hints.status != 2:
            # Check from last_check, see if we may be an implicit vstem
            for i in range(hints.last_checked, index - 1):
                if isinstance(cs.program[i], str):
                    hints.status = 2
                    break
            else:
                # We are an implicit vstem
                hints.has_hint = True
                hints.last_hint = index + 1
                hints.status = 0
        hints.last_checked = index + 1

    def processHint(self, index):
        cs = self.callingStack[-1]
        hints = cs._hints
        hints.has_hint = True
        hints.last_hint = index
        hints.last_checked = index

    def processSubr(self, index, subr):
        cs = self.callingStack[-1]
        hints = cs._hints
        subr_hints = subr._hints

        # Check from last_check, make sure we didn't have
        # any operators.
        if hints.status != 2:
            for i in range(hints.last_checked, index - 1):
                if isinstance(cs.program[i], str):
                    hints.status = 2
                    break
            hints.last_checked = index

        if hints.status != 2:
            if subr_hints.has_hint:
                hints.has_hint = True

                # Decide where to chop off from
                if subr_hints.status == 0:
                    hints.last_hint = index
                else:
                    hints.last_hint = index - 2  # Leave the subr call in

        elif subr_hints.status == 0:
            hints.deletions.append(index)

        hints.status = max(hints.status, subr_hints.status)


def _cs_subset_subroutines(charstring, subrs, gsubrs):
    p = charstring.program
    for i in range(1, len(p)):
        if p[i] == "callsubr":
            assert isinstance(p[i - 1], int)
            p[i - 1] = subrs._used.index(p[i - 1] + subrs._old_bias) - subrs._new_bias
        elif p[i] == "callgsubr":
            assert isinstance(p[i - 1], int)
            p[i - 1] = (
                gsubrs._used.index(p[i - 1] + gsubrs._old_bias) - gsubrs._new_bias
            )


def _cs_drop_hints(charstring):
    hints = charstring._hints

    if hints.deletions:
        p = charstring.program
        for idx in reversed(hints.deletions):
            del p[idx - 2 : idx]

    if hints.has_hint:
        assert not hints.deletions or hints.last_hint <= hints.deletions[0]
        charstring.program = charstring.program[hints.last_hint :]
        if not charstring.program:
            # TODO CFF2 no need for endchar.
            charstring.program.append("endchar")
        if hasattr(charstring, "width"):
            # Insert width back if needed
            if charstring.width != charstring.private.defaultWidthX:
                # For CFF2 charstrings, this should never happen
                assert (
                    charstring.private.defaultWidthX is not None
                ), "CFF2 CharStrings must not have an initial width value"
                charstring.program.insert(
                    0, charstring.width - charstring.private.nominalWidthX
                )

    if hints.has_hintmask:
        i = 0
        p = charstring.program
        while i < len(p):
            if p[i] in ["hintmask", "cntrmask"]:
                assert i + 1 <= len(p)
                del p[i : i + 2]
                continue
            i += 1

    assert len(charstring.program)

    del charstring._hints


def remove_hints(cff, *, removeUnusedSubrs: bool = True):
    for fontname in cff.keys():
        font = cff[fontname]
        cs = font.CharStrings
        # This can be tricky, but doesn't have to. What we do is:
        #
        # - Run all used glyph charstrings and recurse into subroutines,
        # - For each charstring (including subroutines), if it has any
        #   of the hint stem operators, we mark it as such.
        #   Upon returning, for each charstring we note all the
        #   subroutine calls it makes that (recursively) contain a stem,
        # - Dropping hinting then consists of the following two ops:
        #   * Drop the piece of the program in each charstring before the
        #     last call to a stem op or a stem-calling subroutine,
        #   * Drop all hintmask operations.
        # - It's trickier... A hintmask right after hints and a few numbers
        #    will act as an implicit vstemhm. As such, we track whether
        #    we have seen any non-hint operators so far and do the right
        #    thing, recursively... Good luck understanding that :(
        css = set()
        for c in cs.values():
            c.decompile()
            subrs = getattr(c.private, "Subrs", [])
            decompiler = _DehintingT2Decompiler(
                css,
                subrs,
                c.globalSubrs,
                c.private.nominalWidthX,
                c.private.defaultWidthX,
                c.private,
            )
            decompiler.execute(c)
            c.width = decompiler.width
        for charstring in css:
            _cs_drop_hints(charstring)
        del css

        # Drop font-wide hinting values
        all_privs = []
        if hasattr(font, "FDArray"):
            all_privs.extend(fd.Private for fd in font.FDArray)
        else:
            all_privs.append(font.Private)
        for priv in all_privs:
            for k in [
                "BlueValues",
                "OtherBlues",
                "FamilyBlues",
                "FamilyOtherBlues",
                "BlueScale",
                "BlueShift",
                "BlueFuzz",
                "StemSnapH",
                "StemSnapV",
                "StdHW",
                "StdVW",
                "ForceBold",
                "LanguageGroup",
                "ExpansionFactor",
            ]:
                if hasattr(priv, k):
                    setattr(priv, k, None)
    if removeUnusedSubrs:
        remove_unused_subroutines(cff)


def _pd_delete_empty_subrs(private_dict):
    if hasattr(private_dict, "Subrs") and not private_dict.Subrs:
        if "Subrs" in private_dict.rawDict:
            del private_dict.rawDict["Subrs"]
        del private_dict.Subrs


def remove_unused_subroutines(cff):
    for fontname in cff.keys():
        font = cff[fontname]
        cs = font.CharStrings
        # Renumber subroutines to remove unused ones

        # Mark all used subroutines
        for c in cs.values():
            subrs = getattr(c.private, "Subrs", [])
            decompiler = _MarkingT2Decompiler(subrs, c.globalSubrs, c.private)
            decompiler.execute(c)

        all_subrs = [font.GlobalSubrs]
        if hasattr(font, "FDArray"):
            all_subrs.extend(
                fd.Private.Subrs
                for fd in font.FDArray
                if hasattr(fd.Private, "Subrs") and fd.Private.Subrs
            )
        elif hasattr(font.Private, "Subrs") and font.Private.Subrs:
            all_subrs.append(font.Private.Subrs)

        subrs = set(subrs)  # Remove duplicates

        # Prepare
        for subrs in all_subrs:
            if not hasattr(subrs, "_used"):
                subrs._used = set()
            subrs._used = _uniq_sort(subrs._used)
            subrs._old_bias = calcSubrBias(subrs)
            subrs._new_bias = calcSubrBias(subrs._used)

        # Renumber glyph charstrings
        for c in cs.values():
            subrs = getattr(c.private, "Subrs", None)
            _cs_subset_subroutines(c, subrs, font.GlobalSubrs)

        # Renumber subroutines themselves
        for subrs in all_subrs:
            if subrs == font.GlobalSubrs:
                if not hasattr(font, "FDArray") and hasattr(font.Private, "Subrs"):
                    local_subrs = font.Private.Subrs
                elif (
                    hasattr(font, "FDArray")
                    and len(font.FDArray) == 1
                    and hasattr(font.FDArray[0].Private, "Subrs")
                ):
                    # Technically we shouldn't do this. But I've run into fonts that do it.
                    local_subrs = font.FDArray[0].Private.Subrs
                else:
                    local_subrs = None
            else:
                local_subrs = subrs

            subrs.items = [subrs.items[i] for i in subrs._used]
            if hasattr(subrs, "file"):
                del subrs.file
            if hasattr(subrs, "offsets"):
                del subrs.offsets

            for subr in subrs.items:
                _cs_subset_subroutines(subr, local_subrs, font.GlobalSubrs)

        # Delete local SubrsIndex if empty
        if hasattr(font, "FDArray"):
            for fd in font.FDArray:
                _pd_delete_empty_subrs(fd.Private)
        else:
            _pd_delete_empty_subrs(font.Private)

        # Cleanup
        for subrs in all_subrs:
            del subrs._used, subrs._old_bias, subrs._new_bias
